"""
Python Markdown

A Python implementation of John Gruber's Markdown.

Documentation: https://python-markdown.github.io/
GitHub: https://github.com/Python-Markdown/markdown/
PyPI: https://pypi.org/project/Markdown/

Started by Manfred Stienstra (http://www.dwerg.net/).
Maintained for a few years by Yuri Takhteyev (http://www.freewisdom.org).
Currently maintained by Waylan Limberg (https://github.com/waylan),
Dmitry Shachnev (https://github.com/mitya57) and Isaac Muse (https://github.com/facelessuser).

Copyright 2007-2023 The Python Markdown Project (v. 1.7 and later)
Copyright 2004, 2005, 2006 Yuri Takhteyev (v. 0.2-1.6b)
Copyright 2004 Manfred Stienstra (the original version)

License: BSD (see LICENSE.md for details).
"""

from markdown.test_tools import TestCase
from markdown.extensions.tables import TableExtension


class TestTableBlocks(TestCase):

    def test_empty_cells(self):
        """Empty cells (`nbsp`)."""

        text = """
   | Second Header
------------- | -------------
   | Content Cell
Content Cell  |  
"""

        self.assertMarkdownRenders(
            text,
            self.dedent(
                """
                <table>
                <thead>
                <tr>
                <th> </th>
                <th>Second Header</th>
                </tr>
                </thead>
                <tbody>
                <tr>
                <td> </td>
                <td>Content Cell</td>
                </tr>
                <tr>
                <td>Content Cell</td>
                <td> </td>
                </tr>
                </tbody>
                </table>
                """
            ),
            extensions=['tables']
        )

    def test_no_sides(self):
        self.assertMarkdownRenders(
            self.dedent(
                """
                First Header  | Second Header
                ------------- | -------------
                Content Cell  | Content Cell
                Content Cell  | Content Cell
                """
            ),
            self.dedent(
                """
                <table>
                <thead>
                <tr>
                <th>First Header</th>
                <th>Second Header</th>
                </tr>
                </thead>
                <tbody>
                <tr>
                <td>Content Cell</td>
                <td>Content Cell</td>
                </tr>
                <tr>
                <td>Content Cell</td>
                <td>Content Cell</td>
                </tr>
                </tbody>
                </table>
                """
            ),
            extensions=['tables']
        )

    def test_both_sides(self):
        self.assertMarkdownRenders(
            self.dedent(
                """
                | First Header  | Second Header |
                | ------------- | ------------- |
                | Content Cell  | Content Cell  |
                | Content Cell  | Content Cell  |
                """
            ),
            self.dedent(
                """
                <table>
                <thead>
                <tr>
                <th>First Header</th>
                <th>Second Header</th>
                </tr>
                </thead>
                <tbody>
                <tr>
                <td>Content Cell</td>
                <td>Content Cell</td>
                </tr>
                <tr>
                <td>Content Cell</td>
                <td>Content Cell</td>
                </tr>
                </tbody>
                </table>
                """
            ),
            extensions=['tables']
        )

    def test_align_columns(self):
        self.assertMarkdownRenders(
            self.dedent(
                """
                | Item      | Value |
                | :-------- | -----:|
                | Computer  | $1600 |
                | Phone     |   $12 |
                | Pipe      |    $1 |
                """
            ),
            self.dedent(
                """
                <table>
                <thead>
                <tr>
                <th style="text-align: left;">Item</th>
                <th style="text-align: right;">Value</th>
                </tr>
                </thead>
                <tbody>
                <tr>
                <td style="text-align: left;">Computer</td>
                <td style="text-align: right;">$1600</td>
                </tr>
                <tr>
                <td style="text-align: left;">Phone</td>
                <td style="text-align: right;">$12</td>
                </tr>
                <tr>
                <td style="text-align: left;">Pipe</td>
                <td style="text-align: right;">$1</td>
                </tr>
                </tbody>
                </table>
                """
            ),
            extensions=['tables']
        )

    def test_styles_in_tables(self):
        self.assertMarkdownRenders(
            self.dedent(
                """
                | Function name | Description                    |
                | ------------- | ------------------------------ |
                | `help()`      | Display the help window.       |
                | `destroy()`   | **Destroy your computer!**     |
                """
            ),
            self.dedent(
                """
                <table>
                <thead>
                <tr>
                <th>Function name</th>
                <th>Description</th>
                </tr>
                </thead>
                <tbody>
                <tr>
                <td><code>help()</code></td>
                <td>Display the help window.</td>
                </tr>
                <tr>
                <td><code>destroy()</code></td>
                <td><strong>Destroy your computer!</strong></td>
                </tr>
                </tbody>
                </table>
                """
            ),
            extensions=['tables']
        )

    def test_align_three(self):
        self.assertMarkdownRenders(
            self.dedent(
                """
                |foo|bar|baz|
                |:--|:-:|--:|
                |   | Q |   |
                |W  |   |  W|
                """
            ),
            self.dedent(
                """
                <table>
                <thead>
                <tr>
                <th style="text-align: left;">foo</th>
                <th style="text-align: center;">bar</th>
                <th style="text-align: right;">baz</th>
                </tr>
                </thead>
                <tbody>
                <tr>
                <td style="text-align: left;"></td>
                <td style="text-align: center;">Q</td>
                <td style="text-align: right;"></td>
                </tr>
                <tr>
                <td style="text-align: left;">W</td>
                <td style="text-align: center;"></td>
                <td style="text-align: right;">W</td>
                </tr>
                </tbody>
                </table>
                """
            ),
            extensions=['tables']
        )

    def test_three_columns(self):
        self.assertMarkdownRenders(
            self.dedent(
                """
                foo|bar|baz
                ---|---|---
                   | Q |
                 W |   | W
                """
            ),
            self.dedent(
                """
                <table>
                <thead>
                <tr>
                <th>foo</th>
                <th>bar</th>
                <th>baz</th>
                </tr>
                </thead>
                <tbody>
                <tr>
                <td></td>
                <td>Q</td>
                <td></td>
                </tr>
                <tr>
                <td>W</td>
                <td></td>
                <td>W</td>
                </tr>
                </tbody>
                </table>
                """
            ),
            extensions=['tables']
        )

    def test_three_spaces_prefix(self):
        self.assertMarkdownRenders(
            self.dedent(
                """
                Three spaces in front of a table:

                   First Header | Second Header
                   ------------ | -------------
                   Content Cell | Content Cell
                   Content Cell | Content Cell

                   | First Header | Second Header |
                   | ------------ | ------------- |
                   | Content Cell | Content Cell  |
                   | Content Cell | Content Cell  |
                """),
            self.dedent(
                """
                <p>Three spaces in front of a table:</p>
                <table>
                <thead>
                <tr>
                <th>First Header</th>
                <th>Second Header</th>
                </tr>
                </thead>
                <tbody>
                <tr>
                <td>Content Cell</td>
                <td>Content Cell</td>
                </tr>
                <tr>
                <td>Content Cell</td>
                <td>Content Cell</td>
                </tr>
                </tbody>
                </table>
                <table>
                <thead>
                <tr>
                <th>First Header</th>
                <th>Second Header</th>
                </tr>
                </thead>
                <tbody>
                <tr>
                <td>Content Cell</td>
                <td>Content Cell</td>
                </tr>
                <tr>
                <td>Content Cell</td>
                <td>Content Cell</td>
                </tr>
                </tbody>
                </table>
                """
            ),
            extensions=['tables']
        )

    def test_code_block_table(self):
        self.assertMarkdownRenders(
            self.dedent(
                """
                Four spaces is a code block:

                    First Header | Second Header
                    ------------ | -------------
                    Content Cell | Content Cell
                    Content Cell | Content Cell

                | First Header | Second Header |
                | ------------ | ------------- |
                """),
            self.dedent(
                """
                <p>Four spaces is a code block:</p>
                <pre><code>First Header | Second Header
                ------------ | -------------
                Content Cell | Content Cell
                Content Cell | Content Cell
                </code></pre>
                <table>
                <thead>
                <tr>
                <th>First Header</th>
                <th>Second Header</th>
                </tr>
                </thead>
                <tbody>
                <tr>
                <td></td>
                <td></td>
                </tr>
                </tbody>
                </table>
                """
            ),
            extensions=['tables']
        )

    def test_inline_code_blocks(self):
        self.assertMarkdownRenders(
            self.dedent(
                """
                More inline code block tests

                Column 1 | Column 2 | Column 3
                ---------|----------|---------
                word 1   | word 2   | word 3
                word 1   | `word 2` | word 3
                word 1   | \\`word 2 | word 3
                word 1   | `word 2 | word 3
                word 1   | `word |2` | word 3
                words    |`` some | code `` | more words
                words    |``` some | code ``` | more words
                words    |```` some | code ```` | more words
                words    |`` some ` | ` code `` | more words
                words    |``` some ` | ` code ``` | more words
                words    |```` some ` | ` code ```` | more words
                """),
            self.dedent(
                """
                <p>More inline code block tests</p>
                <table>
                <thead>
                <tr>
                <th>Column 1</th>
                <th>Column 2</th>
                <th>Column 3</th>
                </tr>
                </thead>
                <tbody>
                <tr>
                <td>word 1</td>
                <td>word 2</td>
                <td>word 3</td>
                </tr>
                <tr>
                <td>word 1</td>
                <td><code>word 2</code></td>
                <td>word 3</td>
                </tr>
                <tr>
                <td>word 1</td>
                <td>`word 2</td>
                <td>word 3</td>
                </tr>
                <tr>
                <td>word 1</td>
                <td>`word 2</td>
                <td>word 3</td>
                </tr>
                <tr>
                <td>word 1</td>
                <td><code>word |2</code></td>
                <td>word 3</td>
                </tr>
                <tr>
                <td>words</td>
                <td><code>some | code</code></td>
                <td>more words</td>
                </tr>
                <tr>
                <td>words</td>
                <td><code>some | code</code></td>
                <td>more words</td>
                </tr>
                <tr>
                <td>words</td>
                <td><code>some | code</code></td>
                <td>more words</td>
                </tr>
                <tr>
                <td>words</td>
                <td><code>some ` | ` code</code></td>
                <td>more words</td>
                </tr>
                <tr>
                <td>words</td>
                <td><code>some ` | ` code</code></td>
                <td>more words</td>
                </tr>
                <tr>
                <td>words</td>
                <td><code>some ` | ` code</code></td>
                <td>more words</td>
                </tr>
                </tbody>
                </table>
                """
            ),
            extensions=['tables']
        )

    def test_issue_440(self):
        self.assertMarkdownRenders(
            self.dedent(
                """
                A test for issue #440:

                foo | bar
                --- | ---
                foo | (`bar`) and `baz`.
                """),
            self.dedent(
                """
                <p>A test for issue #440:</p>
                <table>
                <thead>
                <tr>
                <th>foo</th>
                <th>bar</th>
                </tr>
                </thead>
                <tbody>
                <tr>
                <td>foo</td>
                <td>(<code>bar</code>) and <code>baz</code>.</td>
                </tr>
                </tbody>
                </table>
                """
            ),
            extensions=['tables']
        )

    def test_lists_not_tables(self):
        self.assertMarkdownRenders(
            self.dedent(
                """
                Lists are not tables

                 - this | should | not
                 - be | a | table
                """),
            self.dedent(
                """
                <p>Lists are not tables</p>
                <ul>
                <li>this | should | not</li>
                <li>be | a | table</li>
                </ul>
                """
            ),
            extensions=['tables']
        )

    def test_issue_449(self):
        self.assertMarkdownRenders(
            self.dedent(
                r"""
                Add tests for issue #449

                Odd backticks | Even backticks
                ------------ | -------------
                ``[!\"\#$%&'()*+,\-./:;<=>?@\[\\\]^_`{|}~]`` | ``[!\"\#$%&'()*+,\-./:;<=>?@\[\\\]^`_`{|}~]``

                Escapes | More Escapes
                ------- | ------
                `` `\`` | `\`

                Only the first backtick can be escaped

                Escaped | Bacticks
                ------- | ------
                \`` \`  | \`\`

                Test escaped pipes

                Column 1 | Column 2
                -------- | --------
                `|` \|   | Pipes are okay in code and escaped. \|

                | Column 1 | Column 2 |
                | -------- | -------- |
                | row1     | row1    \|
                | row2     | row2     |

                Test header escapes

                | `` `\`` \| | `\` \|
                | ---------- | ---- |
                | row1       | row1 |
                | row2       | row2 |

                Escaped pipes in format row should not be a table

                | Column1   | Column2 |
                | ------- \|| ------- |
                | row1      | row1    |
                | row2      | row2    |

                Test escaped code in Table

                Should not be code | Should be code
                ------------------ | --------------
                \`Not code\`       | \\`code`
                \\\`Not code\\\`   | \\\\`code`
                """),
            self.dedent(
                """
                <p>Add tests for issue #449</p>
                <table>
                <thead>
                <tr>
                <th>Odd backticks</th>
                <th>Even backticks</th>
                </tr>
                </thead>
                <tbody>
                <tr>
                <td><code>[!\\"\\#$%&amp;'()*+,\\-./:;&lt;=&gt;?@\\[\\\\\\]^_`{|}~]</code></td>
                <td><code>[!\\"\\#$%&amp;'()*+,\\-./:;&lt;=&gt;?@\\[\\\\\\]^`_`{|}~]</code></td>
                </tr>
                </tbody>
                </table>
                <table>
                <thead>
                <tr>
                <th>Escapes</th>
                <th>More Escapes</th>
                </tr>
                </thead>
                <tbody>
                <tr>
                <td><code>`\\</code></td>
                <td><code>\\</code></td>
                </tr>
                </tbody>
                </table>
                <p>Only the first backtick can be escaped</p>
                <table>
                <thead>
                <tr>
                <th>Escaped</th>
                <th>Bacticks</th>
                </tr>
                </thead>
                <tbody>
                <tr>
                <td>`<code>\\</code></td>
                <td>``</td>
                </tr>
                </tbody>
                </table>
                <p>Test escaped pipes</p>
                <table>
                <thead>
                <tr>
                <th>Column 1</th>
                <th>Column 2</th>
                </tr>
                </thead>
                <tbody>
                <tr>
                <td><code>|</code> |</td>
                <td>Pipes are okay in code and escaped. |</td>
                </tr>
                </tbody>
                </table>
                <table>
                <thead>
                <tr>
                <th>Column 1</th>
                <th>Column 2</th>
                </tr>
                </thead>
                <tbody>
                <tr>
                <td>row1</td>
                <td>row1    |</td>
                </tr>
                <tr>
                <td>row2</td>
                <td>row2</td>
                </tr>
                </tbody>
                </table>
                <p>Test header escapes</p>
                <table>
                <thead>
                <tr>
                <th><code>`\\</code> |</th>
                <th><code>\\</code> |</th>
                </tr>
                </thead>
                <tbody>
                <tr>
                <td>row1</td>
                <td>row1</td>
                </tr>
                <tr>
                <td>row2</td>
                <td>row2</td>
                </tr>
                </tbody>
                </table>
                <p>Escaped pipes in format row should not be a table</p>
                <p>| Column1   | Column2 |
                | ------- || ------- |
                | row1      | row1    |
                | row2      | row2    |</p>
                <p>Test escaped code in Table</p>
                <table>
                <thead>
                <tr>
                <th>Should not be code</th>
                <th>Should be code</th>
                </tr>
                </thead>
                <tbody>
                <tr>
                <td>`Not code`</td>
                <td>\\<code>code</code></td>
                </tr>
                <tr>
                <td>\\`Not code\\`</td>
                <td>\\\\<code>code</code></td>
                </tr>
                </tbody>
                </table>
                """
            ),
            extensions=['tables']
        )

    def test_single_column_tables(self):
        self.assertMarkdownRenders(
            self.dedent(
                """
                Single column tables

                | Is a Table |
                | ---------- |

                | Is a Table
                | ----------

                Is a Table |
                ---------- |

                | Is a Table |
                | ---------- |
                | row        |

                | Is a Table
                | ----------
                | row

                Is a Table |
                ---------- |
                row        |

                | Is not a Table
                --------------
                | row

                Is not a Table |
                --------------
                row            |

                | Is not a Table
                | --------------
                row

                Is not a Table |
                -------------- |
                row
                """),
            self.dedent(
                """
                <p>Single column tables</p>
                <table>
                <thead>
                <tr>
                <th>Is a Table</th>
                </tr>
                </thead>
                <tbody>
                <tr>
                <td></td>
                </tr>
                </tbody>
                </table>
                <table>
                <thead>
                <tr>
                <th>Is a Table</th>
                </tr>
                </thead>
                <tbody>
                <tr>
                <td></td>
                </tr>
                </tbody>
                </table>
                <table>
                <thead>
                <tr>
                <th>Is a Table</th>
                </tr>
                </thead>
                <tbody>
                <tr>
                <td></td>
                </tr>
                </tbody>
                </table>
                <table>
                <thead>
                <tr>
                <th>Is a Table</th>
                </tr>
                </thead>
                <tbody>
                <tr>
                <td>row</td>
                </tr>
                </tbody>
                </table>
                <table>
                <thead>
                <tr>
                <th>Is a Table</th>
                </tr>
                </thead>
                <tbody>
                <tr>
                <td>row</td>
                </tr>
                </tbody>
                </table>
                <table>
                <thead>
                <tr>
                <th>Is a Table</th>
                </tr>
                </thead>
                <tbody>
                <tr>
                <td>row</td>
                </tr>
                </tbody>
                </table>
                <h2>| Is not a Table</h2>
                <p>| row</p>
                <h2>Is not a Table |</h2>
                <p>row            |</p>
                <p>| Is not a Table
                | --------------
                row</p>
                <p>Is not a Table |
                -------------- |
                row</p>
                """
            ),
            extensions=['tables']
        )

    def test_align_columns_legacy(self):
        self.assertMarkdownRenders(
            self.dedent(
                """
                | Item      | Value |
                | :-------- | -----:|
                | Computer  | $1600 |
                | Phone     |   $12 |
                | Pipe      |    $1 |
                """
            ),
            self.dedent(
                """
                <table>
                <thead>
                <tr>
                <th align="left">Item</th>
                <th align="right">Value</th>
                </tr>
                </thead>
                <tbody>
                <tr>
                <td align="left">Computer</td>
                <td align="right">$1600</td>
                </tr>
                <tr>
                <td align="left">Phone</td>
                <td align="right">$12</td>
                </tr>
                <tr>
                <td align="left">Pipe</td>
                <td align="right">$1</td>
                </tr>
                </tbody>
                </table>
                """
            ),
            extensions=[TableExtension(use_align_attribute=True)]
        )

    def test_align_three_legacy(self):
        self.assertMarkdownRenders(
            self.dedent(
                """
                |foo|bar|baz|
                |:--|:-:|--:|
                |   | Q |   |
                |W  |   |  W|
                """
            ),
            self.dedent(
                """
                <table>
                <thead>
                <tr>
                <th align="left">foo</th>
                <th align="center">bar</th>
                <th align="right">baz</th>
                </tr>
                </thead>
                <tbody>
                <tr>
                <td align="left"></td>
                <td align="center">Q</td>
                <td align="right"></td>
                </tr>
                <tr>
                <td align="left">W</td>
                <td align="center"></td>
                <td align="right">W</td>
                </tr>
                </tbody>
                </table>
                """
            ),
            extensions=[TableExtension(use_align_attribute=True)]
        )



"""
Python Markdown

A Python implementation of John Gruber's Markdown.

Documentation: https://python-markdown.github.io/
GitHub: https://github.com/Python-Markdown/markdown/
PyPI: https://pypi.org/project/Markdown/

Started by Manfred Stienstra (http://www.dwerg.net/).
Maintained for a few years by Yuri Takhteyev (http://www.freewisdom.org).
Currently maintained by Waylan Limberg (https://github.com/waylan),
Dmitry Shachnev (https://github.com/mitya57) and Isaac Muse (https://github.com/facelessuser).

Copyright 2007-2019 The Python Markdown Project (v. 1.7 and later)
Copyright 2004, 2005, 2006 Yuri Takhteyev (v. 0.2-1.6b)
Copyright 2004 Manfred Stienstra (the original version)

License: BSD (see LICENSE.md for details).
"""

from markdown.test_tools import TestCase
from markdown import Markdown
from markdown.extensions.toc import TocExtension, strip_tags, unique
from markdown.extensions.nl2br import Nl2BrExtension


class TestTOC(TestCase):
    maxDiff = None
    default_kwargs = {
        'extensions': [TocExtension()]
    }

    def testTOCMarker(self):
        self.assertMarkdownRenders(
            self.dedent(
                '''
                [TOC]

                # Header 1

                ## Header 2
                '''
            ),
            '<div class="toc">\n'
              '<ul>\n'                                             # noqa
                '<li><a href="#header-1">Header 1</a>'             # noqa
                  '<ul>\n'                                         # noqa
                    '<li><a href="#header-2">Header 2</a></li>\n'  # noqa
                  '</ul>\n'                                        # noqa
                '</li>\n'                                          # noqa
              '</ul>\n'                                            # noqa
            '</div>\n'
            '<h1 id="header-1">Header 1</h1>\n'
            '<h2 id="header-2">Header 2</h2>'
        )

    def testNoTOCMarker(self):
        self.assertMarkdownRenders(
            self.dedent(
                '''
                # Header 1

                ## Header 2
                '''
            ),
            self.dedent(
                '''
                <h1 id="header-1">Header 1</h1>
                <h2 id="header-2">Header 2</h2>
                '''
            ),
            expected_attrs={
                'toc': (
                    '<div class="toc">\n'
                      '<ul>\n'                                             # noqa
                        '<li><a href="#header-1">Header 1</a>'             # noqa
                          '<ul>\n'                                         # noqa
                            '<li><a href="#header-2">Header 2</a></li>\n'  # noqa
                          '</ul>\n'                                        # noqa
                        '</li>\n'                                          # noqa
                      '</ul>\n'                                            # noqa
                    '</div>\n'
                )
            }
        )

    def testAlternateTOCMarker(self):
        self.assertMarkdownRenders(
            self.dedent(
                '''
                {{marker}}

                # Header 1

                ## Header 2
                '''
            ),
            '<div class="toc">\n'
              '<ul>\n'                                             # noqa
                '<li><a href="#header-1">Header 1</a>'             # noqa
                  '<ul>\n'                                         # noqa
                    '<li><a href="#header-2">Header 2</a></li>\n'  # noqa
                  '</ul>\n'                                        # noqa
                '</li>\n'                                          # noqa
              '</ul>\n'                                            # noqa
            '</div>\n'
            '<h1 id="header-1">Header 1</h1>\n'
            '<h2 id="header-2">Header 2</h2>',
            extensions=[TocExtension(marker='{{marker}}')]
        )

    def testDisabledTOCMarker(self):
        self.assertMarkdownRenders(
            self.dedent(
                '''
                [TOC]

                # Header 1

                ## Header 2
                '''
            ),
            self.dedent(
                '''
                <p>[TOC]</p>
                <h1 id="header-1">Header 1</h1>
                <h2 id="header-2">Header 2</h2>
                '''
            ),
            expected_attrs={
                'toc': (
                    '<div class="toc">\n'
                      '<ul>\n'                                             # noqa
                        '<li><a href="#header-1">Header 1</a>'             # noqa
                          '<ul>\n'                                         # noqa
                            '<li><a href="#header-2">Header 2</a></li>\n'  # noqa
                          '</ul>\n'                                        # noqa
                        '</li>\n'                                          # noqa
                      '</ul>\n'                                            # noqa
                    '</div>\n'
                )
            },
            extensions=[TocExtension(marker='')]
        )

    def testTOCReset(self):
        md = Markdown(extensions=[TocExtension()])
        self.assertEqual(md.toc, '')
        self.assertEqual(md.toc_tokens, [])
        md.convert('# Header 1')
        self.assertEqual('<div class="toc">', md.toc[:17])
        self.assertEqual(len(md.toc_tokens), 1)
        md.reset()
        self.assertEqual(md.toc, '')
        self.assertEqual(md.toc_tokens, [])

    def testUniqueIds(self):
        self.assertMarkdownRenders(
            self.dedent(
                '''
                #Header
                #Header
                #Header
                '''
            ),
            self.dedent(
                '''
                <h1 id="header">Header</h1>
                <h1 id="header_1">Header</h1>
                <h1 id="header_2">Header</h1>
                '''
            ),
            expected_attrs={
                'toc': (
                    '<div class="toc">\n'
                      '<ul>\n'                                       # noqa
                        '<li><a href="#header">Header</a></li>\n'    # noqa
                        '<li><a href="#header_1">Header</a></li>\n'  # noqa
                        '<li><a href="#header_2">Header</a></li>\n'  # noqa
                      '</ul>\n'                                      # noqa
                    '</div>\n'
                ),
                'toc_tokens': [
                    {
                        'level': 1,
                        'id': 'header',
                        'name': 'Header',
                        'html': 'Header',
                        'data-toc-label': '',
                        'children': []
                    },
                    {
                        'level': 1,
                        'id': 'header_1',
                        'name': 'Header',
                        'html': 'Header',
                        'data-toc-label': '',
                        'children': []
                    },
                    {
                        'level': 1,
                        'id': 'header_2',
                        'name': 'Header',
                        'html': 'Header',
                        'data-toc-label': '',
                        'children': []
                    },
                ]
            }
        )

    def testHtmlEntitiesInTOC(self):
        self.assertMarkdownRenders(
            '# Foo &amp; bar',
            '<h1 id="foo-bar">Foo &amp; bar</h1>',
            expected_attrs={
                'toc': (
                    '<div class="toc">\n'
                      '<ul>\n'                                             # noqa
                        '<li><a href="#foo-bar">Foo &amp; bar</a></li>\n'  # noqa
                      '</ul>\n'                                            # noqa
                    '</div>\n'
                ),
                'toc_tokens': [{
                    'level': 1,
                    'id': 'foo-bar',
                    'name': 'Foo &amp; bar',
                    'html': 'Foo &amp; bar',
                    'data-toc-label': '',
                    'children': []
                }]
            }
        )

    def testHtmlSpecialCharsInTOC(self):
        self.assertMarkdownRenders(
            '# Foo > & bar',
            '<h1 id="foo-bar">Foo &gt; &amp; bar</h1>',
            expected_attrs={
                'toc': (
                    '<div class="toc">\n'
                      '<ul>\n'                                                  # noqa
                        '<li><a href="#foo-bar">Foo &gt; &amp; bar</a></li>\n'  # noqa
                      '</ul>\n'                                                 # noqa
                    '</div>\n'
                ),
                'toc_tokens': [{
                    'level': 1,
                    'id': 'foo-bar',
                    'name': 'Foo &gt; &amp; bar',
                    'html': 'Foo &gt; &amp; bar',
                    'data-toc-label': '',
                    'children': []
                }]
            }
        )

    def testRawHtmlInTOC(self):
        self.assertMarkdownRenders(
            '# Foo <b>Bar</b> Baz.',
            '<h1 id="foo-bar-baz">Foo <b>Bar</b> Baz.</h1>',
            expected_attrs={
                'toc': (
                    '<div class="toc">\n'
                      '<ul>\n'                                                # noqa
                        '<li><a href="#foo-bar-baz">Foo Bar Baz.</a></li>\n'  # noqa
                      '</ul>\n'                                               # noqa
                    '</div>\n'
                ),
                'toc_tokens': [{
                    'level': 1,
                    'id': 'foo-bar-baz',
                    'name': 'Foo Bar Baz.',
                    'html': 'Foo <b>Bar</b> Baz.',
                    'data-toc-label': '',
                    'children': []
                }]
            }
        )

    def testTOCBaseLevel(self):
        self.assertMarkdownRenders(
            self.dedent(
                '''
                # Some Header
                ## Next Level
                ### Too High
                '''
            ),
            self.dedent(
                '''
                <h5 id="some-header">Some Header</h5>
                <h6 id="next-level">Next Level</h6>
                <h6 id="too-high">Too High</h6>
                '''
            ),
            expected_attrs={
                'toc': (
                    '<div class="toc">\n'
                      '<ul>\n'                                                 # noqa
                        '<li><a href="#some-header">Some Header</a>'           # noqa
                          '<ul>\n'                                             # noqa
                            '<li><a href="#next-level">Next Level</a></li>\n'  # noqa
                            '<li><a href="#too-high">Too High</a></li>\n'      # noqa
                          '</ul>\n'                                            # noqa
                        '</li>\n'                                              # noqa
                      '</ul>\n'                                                # noqa
                    '</div>\n'
                ),
                'toc_tokens': [{
                    'level': 5,
                    'id': 'some-header',
                    'name': 'Some Header',
                    'html': 'Some Header',
                    'data-toc-label': '',
                    'children': [
                        {
                            'level': 6,
                            'id': 'next-level',
                            'name': 'Next Level',
                            'html': 'Next Level',
                            'data-toc-label': '',
                            'children': []
                        },
                        {
                            'level': 6,
                            'id': 'too-high',
                            'name': 'Too High',
                            'html': 'Too High',
                            'data-toc-label': '',
                            'children': []
                        }
                    ]
                }]
            },
            extensions=[TocExtension(baselevel=5)]
        )

    def testHeaderInlineMarkup(self):
        self.assertMarkdownRenders(
            '#Some *Header* with [markup](http://example.com).',
            '<h1 id="some-header-with-markup">Some <em>Header</em> with '
            '<a href="http://example.com">markup</a>.</h1>',
            expected_attrs={
                'toc': (
                    '<div class="toc">\n'
                      '<ul>\n'                                     # noqa
                        '<li><a href="#some-header-with-markup">'  # noqa
                          'Some Header with markup.</a></li>\n'    # noqa
                      '</ul>\n'                                    # noqa
                    '</div>\n'
                ),
                'toc_tokens': [{
                    'level': 1,
                    'id': 'some-header-with-markup',
                    'name': 'Some Header with markup.',
                    'html': 'Some <em>Header</em> with <a href="http://example.com">markup</a>.',
                    'data-toc-label': '',
                    'children': []
                }]
            }
        )

    def testTOCTitle(self):
        self.assertMarkdownRenders(
            self.dedent(
                '''
                # Header 1

                ## Header 2
                '''
            ),
            self.dedent(
                '''
                <h1 id="header-1">Header 1</h1>
                <h2 id="header-2">Header 2</h2>
                '''
            ),
            expected_attrs={
                'toc': (
                    '<div class="toc"><span class="toctitle">Table of Contents</span>'
                      '<ul>\n'                                             # noqa
                        '<li><a href="#header-1">Header 1</a>'             # noqa
                          '<ul>\n'                                         # noqa
                            '<li><a href="#header-2">Header 2</a></li>\n'  # noqa
                          '</ul>\n'                                        # noqa
                        '</li>\n'                                          # noqa
                      '</ul>\n'                                            # noqa
                    '</div>\n'
                )
            },
            extensions=[TocExtension(title='Table of Contents')]
        )

    def testTOCUniqueFunc(self):
        ids = {'foo'}
        self.assertEqual(unique('foo', ids), 'foo_1')
        self.assertEqual(ids, {'foo', 'foo_1'})

    def testTocInHeaders(self):
        self.assertMarkdownRenders(
            self.dedent(
                '''
                [TOC]
                #[TOC]
                '''
            ),
            '<div class="toc">\n'                       # noqa
              '<ul>\n'                                  # noqa
                '<li><a href="#toc">[TOC]</a></li>\n'   # noqa
              '</ul>\n'                                 # noqa
            '</div>\n'                                  # noqa
            '<h1 id="toc">[TOC]</h1>'                   # noqa
        )

        self.assertMarkdownRenders(
            self.dedent(
                '''
                #[TOC]
                [TOC]
                '''
            ),
            '<h1 id="toc">[TOC]</h1>\n'                 # noqa
            '<div class="toc">\n'                       # noqa
              '<ul>\n'                                  # noqa
                '<li><a href="#toc">[TOC]</a></li>\n'   # noqa
              '</ul>\n'                                 # noqa
            '</div>'                                    # noqa
        )

        self.assertMarkdownRenders(
            self.dedent(
                '''
                [TOC]
                # *[TOC]*
                '''
            ),
            '<div class="toc">\n'                       # noqa
              '<ul>\n'                                  # noqa
                '<li><a href="#toc">[TOC]</a></li>\n'   # noqa
              '</ul>\n'                                 # noqa
            '</div>\n'                                  # noqa
            '<h1 id="toc"><em>[TOC]</em></h1>'          # noqa
        )

    def testTOCPermalink(self):
        self.assertMarkdownRenders(
            self.dedent(
                '''
                # Hd 1

                ## Hd 2
                '''
            ),
            '<h1 id="hd-1">'
                'Hd 1'                                            # noqa
                '<a class="headerlink" href="#hd-1" title="PL">'  # noqa
                    '&para;'                                      # noqa
                '</a>'                                            # noqa
            '</h1>\n'
            '<h2 id="hd-2">'
                'Hd 2'                                            # noqa
                '<a class="headerlink" href="#hd-2" title="PL">'  # noqa
                    '&para;'                                      # noqa
                '</a>'                                            # noqa
            '</h2>',
            extensions=[TocExtension(permalink=True, permalink_title="PL")]
        )

    def testTOCPermalinkLeading(self):
        self.assertMarkdownRenders(
            self.dedent(
                '''
                # Hd 1
                ## Hd 2
                '''
            ),
            '<h1 id="hd-1">'
                '<a class="headerlink" href="#hd-1" title="PL">'  # noqa
                    '&para;'                                      # noqa
                '</a>'                                            # noqa
                'Hd 1'                                            # noqa
            '</h1>\n'
            '<h2 id="hd-2">'
                '<a class="headerlink" href="#hd-2" title="PL">'  # noqa
                    '&para;'                                      # noqa
                '</a>'                                            # noqa
                'Hd 2'                                            # noqa
            '</h2>',
            extensions=[TocExtension(permalink=True, permalink_title="PL", permalink_leading=True)]
        )

    def testTOCInlineMarkupPermalink(self):
        self.assertMarkdownRenders(
            '# Code `in` hd',
            '<h1 id="code-in-hd">'
                'Code <code>in</code> hd'                               # noqa
                '<a class="headerlink" href="#code-in-hd" title="PL">'  # noqa
                    '&para;'                                            # noqa
                '</a>'                                                  # noqa
            '</h1>',
            extensions=[TocExtension(permalink=True, permalink_title="PL")]
        )

    def testTOCInlineMarkupPermalinkLeading(self):
        self.assertMarkdownRenders(
            '# Code `in` hd',
            '<h1 id="code-in-hd">'
                '<a class="headerlink" href="#code-in-hd" title="PL">'  # noqa
                    '&para;'                                            # noqa
                '</a>'                                                  # noqa
                'Code <code>in</code> hd'                               # noqa
            '</h1>',
            extensions=[TocExtension(permalink=True, permalink_title="PL", permalink_leading=True)]
        )

    def testAnchorLink(self):
        self.assertMarkdownRenders(
            self.dedent(
                '''
                # Header 1

                ## Header *2*
                '''
            ),
            self.dedent(
                '''
                <h1 id="header-1"><a class="toclink" href="#header-1">Header 1</a></h1>
                <h2 id="header-2"><a class="toclink" href="#header-2">Header <em>2</em></a></h2>
                '''
            ),
            extensions=[TocExtension(anchorlink=True)]
        )

    def testAnchorLinkWithSingleInlineCode(self):
        self.assertMarkdownRenders(
            '# This is `code`.',
            '<h1 id="this-is-code">'                        # noqa
                '<a class="toclink" href="#this-is-code">'  # noqa
                    'This is <code>code</code>.'            # noqa
                '</a>'                                      # noqa
            '</h1>',                                        # noqa
            extensions=[TocExtension(anchorlink=True)]
        )

    def testAnchorLinkWithDoubleInlineCode(self):
        self.assertMarkdownRenders(
            '# This is `code` and `this` too.',
            '<h1 id="this-is-code-and-this-too">'                           # noqa
                '<a class="toclink" href="#this-is-code-and-this-too">'     # noqa
                    'This is <code>code</code> and <code>this</code> too.'  # noqa
                '</a>'                                                      # noqa
            '</h1>',                                                        # noqa
            extensions=[TocExtension(anchorlink=True)]
        )

    def testPermalink(self):
        self.assertMarkdownRenders(
            '# Header',
            '<h1 id="header">'                                                            # noqa
                'Header'                                                                  # noqa
                '<a class="headerlink" href="#header" title="Permanent link">&para;</a>'  # noqa
            '</h1>',                                                                      # noqa
            extensions=[TocExtension(permalink=True)]
        )

    def testPermalinkWithSingleInlineCode(self):
        self.assertMarkdownRenders(
            '# This is `code`.',
            '<h1 id="this-is-code">'                                                            # noqa
                'This is <code>code</code>.'                                                    # noqa
                '<a class="headerlink" href="#this-is-code" title="Permanent link">&para;</a>'  # noqa
            '</h1>',                                                                            # noqa
            extensions=[TocExtension(permalink=True)]
        )

    def testPermalinkWithDoubleInlineCode(self):
        self.assertMarkdownRenders(
            '# This is `code` and `this` too.',
            '<h1 id="this-is-code-and-this-too">'                                                            # noqa
                'This is <code>code</code> and <code>this</code> too.'                                       # noqa
                '<a class="headerlink" href="#this-is-code-and-this-too" title="Permanent link">&para;</a>'  # noqa
            '</h1>',                                                                                         # noqa
            extensions=[TocExtension(permalink=True)]
        )

    def testMinMaxLevel(self):
        self.assertMarkdownRenders(
            self.dedent(
                '''
                # Header 1 not in TOC

                ## Header 2 not in TOC

                ### Header 3

                #### Header 4

                ##### Header 5 not in TOC
                '''
            ),
            self.dedent(
                '''
                <h1 id="header-1-not-in-toc">Header 1 not in TOC</h1>
                <h2 id="header-2-not-in-toc">Header 2 not in TOC</h2>
                <h3 id="header-3">Header 3</h3>
                <h4 id="header-4">Header 4</h4>
                <h5 id="header-5-not-in-toc">Header 5 not in TOC</h5>
                '''
            ),
            expected_attrs={
                'toc': (
                    '<div class="toc">\n'
                      '<ul>\n'                                             # noqa
                        '<li><a href="#header-3">Header 3</a>'             # noqa
                          '<ul>\n'                                         # noqa
                            '<li><a href="#header-4">Header 4</a></li>\n'  # noqa
                          '</ul>\n'                                        # noqa
                        '</li>\n'                                          # noqa
                      '</ul>\n'                                            # noqa
                    '</div>\n'                                             # noqa
                ),
                'toc_tokens': [
                    {
                        'level': 3,
                        'id': 'header-3',
                        'name': 'Header 3',
                        'html': 'Header 3',
                        'data-toc-label': '',
                        'children': [
                            {
                                'level': 4,
                                'id': 'header-4',
                                'name': 'Header 4',
                                'html': 'Header 4',
                                'data-toc-label': '',
                                'children': []
                            }
                        ]
                    }
                ]
            },
            extensions=[TocExtension(toc_depth='3-4')]
        )

    def testMaxLevel(self):
        self.assertMarkdownRenders(
            self.dedent(
                '''
                # Header 1

                ## Header 2

                ### Header 3 not in TOC
                '''
            ),
            self.dedent(
                '''
                <h1 id="header-1">Header 1</h1>
                <h2 id="header-2">Header 2</h2>
                <h3 id="header-3-not-in-toc">Header 3 not in TOC</h3>
                '''
            ),
            expected_attrs={
                'toc': (
                    '<div class="toc">\n'
                      '<ul>\n'                                             # noqa
                        '<li><a href="#header-1">Header 1</a>'             # noqa
                          '<ul>\n'                                         # noqa
                            '<li><a href="#header-2">Header 2</a></li>\n'  # noqa
                          '</ul>\n'                                        # noqa
                        '</li>\n'                                          # noqa
                      '</ul>\n'                                            # noqa
                    '</div>\n'                                             # noqa
                ),
                'toc_tokens': [
                    {
                        'level': 1,
                        'id': 'header-1',
                        'name': 'Header 1',
                        'html': 'Header 1',
                        'data-toc-label': '',
                        'children': [
                            {
                                'level': 2,
                                'id': 'header-2',
                                'name': 'Header 2',
                                'html': 'Header 2',
                                'data-toc-label': '',
                                'children': []
                            }
                        ]
                    }
                ]
            },
            extensions=[TocExtension(toc_depth=2)]
        )

    def testMinMaxLevelwithAnchorLink(self):
        self.assertMarkdownRenders(
            self.dedent(
                '''
                # Header 1 not in TOC

                ## Header 2 not in TOC

                ### Header 3

                #### Header 4

                ##### Header 5 not in TOC
                '''
            ),
            '<h1 id="header-1-not-in-toc">'                                                      # noqa
                '<a class="toclink" href="#header-1-not-in-toc">Header 1 not in TOC</a></h1>\n'  # noqa
            '<h2 id="header-2-not-in-toc">'                                                      # noqa
                '<a class="toclink" href="#header-2-not-in-toc">Header 2 not in TOC</a></h2>\n'  # noqa
            '<h3 id="header-3">'                                                                 # noqa
                '<a class="toclink" href="#header-3">Header 3</a></h3>\n'                        # noqa
            '<h4 id="header-4">'                                                                 # noqa
                '<a class="toclink" href="#header-4">Header 4</a></h4>\n'                        # noqa
            '<h5 id="header-5-not-in-toc">'                                                      # noqa
                '<a class="toclink" href="#header-5-not-in-toc">Header 5 not in TOC</a></h5>',   # noqa
            expected_attrs={
                'toc': (
                    '<div class="toc">\n'
                      '<ul>\n'                                             # noqa
                        '<li><a href="#header-3">Header 3</a>'             # noqa
                          '<ul>\n'                                         # noqa
                            '<li><a href="#header-4">Header 4</a></li>\n'  # noqa
                          '</ul>\n'                                        # noqa
                        '</li>\n'                                          # noqa
                      '</ul>\n'                                            # noqa
                    '</div>\n'                                             # noqa
                ),
                'toc_tokens': [
                    {
                        'level': 3,
                        'id': 'header-3',
                        'name': 'Header 3',
                        'html': 'Header 3',
                        'data-toc-label': '',
                        'children': [
                            {
                                'level': 4,
                                'id': 'header-4',
                                'name': 'Header 4',
                                'html': 'Header 4',
                                'data-toc-label': '',
                                'children': []
                            }
                        ]
                    }
                ]
            },
            extensions=[TocExtension(toc_depth='3-4', anchorlink=True)]
        )

    def testMinMaxLevelwithPermalink(self):
        self.assertMarkdownRenders(
            self.dedent(
                '''
                # Header 1 not in TOC

                ## Header 2 not in TOC

                ### Header 3

                #### Header 4

                ##### Header 5 not in TOC
                '''
            ),
            '<h1 id="header-1-not-in-toc">Header 1 not in TOC'                                                # noqa
                '<a class="headerlink" href="#header-1-not-in-toc" title="Permanent link">&para;</a></h1>\n'  # noqa
            '<h2 id="header-2-not-in-toc">Header 2 not in TOC'                                                # noqa
                '<a class="headerlink" href="#header-2-not-in-toc" title="Permanent link">&para;</a></h2>\n'  # noqa
            '<h3 id="header-3">Header 3'                                                                      # noqa
                '<a class="headerlink" href="#header-3" title="Permanent link">&para;</a></h3>\n'             # noqa
            '<h4 id="header-4">Header 4'                                                                      # noqa
                '<a class="headerlink" href="#header-4" title="Permanent link">&para;</a></h4>\n'             # noqa
            '<h5 id="header-5-not-in-toc">Header 5 not in TOC'                                                # noqa
                '<a class="headerlink" href="#header-5-not-in-toc" title="Permanent link">&para;</a></h5>',   # noqa
            expected_attrs={
                'toc': (
                    '<div class="toc">\n'
                      '<ul>\n'                                             # noqa
                        '<li><a href="#header-3">Header 3</a>'             # noqa
                          '<ul>\n'                                         # noqa
                            '<li><a href="#header-4">Header 4</a></li>\n'  # noqa
                          '</ul>\n'                                        # noqa
                        '</li>\n'                                          # noqa
                      '</ul>\n'                                            # noqa
                    '</div>\n'                                             # noqa
                ),
                'toc_tokens': [
                    {
                        'level': 3,
                        'id': 'header-3',
                        'name': 'Header 3',
                        'html': 'Header 3',
                        'data-toc-label': '',
                        'children': [
                            {
                                'level': 4,
                                'id': 'header-4',
                                'name': 'Header 4',
                                'html': 'Header 4',
                                'data-toc-label': '',
                                'children': []
                            }
                        ]
                    }
                ]
            },
            extensions=[TocExtension(toc_depth='3-4', permalink=True)]
        )

    def testMinMaxLevelwithBaseLevel(self):
        self.assertMarkdownRenders(
            self.dedent(
                '''
                # First Header

                ## Second Level

                ### Third Level

                #### Forth Level
                '''
            ),
            self.dedent(
                '''
                <h3 id="first-header">First Header</h3>
                <h4 id="second-level">Second Level</h4>
                <h5 id="third-level">Third Level</h5>
                <h6 id="forth-level">Forth Level</h6>
                '''
            ),
            expected_attrs={
                'toc': (
                    '<div class="toc">\n'
                      '<ul>\n'                                                  # noqa
                        '<li><a href="#second-level">Second Level</a>'          # noqa
                          '<ul>\n'                                              # noqa
                            '<li><a href="#third-level">Third Level</a></li>\n' # noqa
                          '</ul>\n'                                             # noqa
                        '</li>\n'                                               # noqa
                      '</ul>\n'                                                 # noqa
                    '</div>\n'                                                  # noqa
                ),
                'toc_tokens': [
                    {
                        'level': 4,
                        'id': 'second-level',
                        'name': 'Second Level',
                        'html': 'Second Level',
                        'data-toc-label': '',
                        'children': [
                            {
                                'level': 5,
                                'id': 'third-level',
                                'name': 'Third Level',
                                'html': 'Third Level',
                                'data-toc-label': '',
                                'children': []
                            }
                        ]
                    }
                ]
            },
            extensions=[TocExtension(toc_depth='4-5', baselevel=3)]
        )

    def testMaxLevelwithBaseLevel(self):
        self.assertMarkdownRenders(
            self.dedent(
                '''
                # Some Header

                ## Next Level

                ### Too High
                '''
            ),
            self.dedent(
                '''
                <h2 id="some-header">Some Header</h2>
                <h3 id="next-level">Next Level</h3>
                <h4 id="too-high">Too High</h4>
                '''
            ),
            expected_attrs={
                'toc': (
                    '<div class="toc">\n'
                      '<ul>\n'                                                 # noqa
                        '<li><a href="#some-header">Some Header</a>'           # noqa
                          '<ul>\n'                                             # noqa
                            '<li><a href="#next-level">Next Level</a></li>\n'  # noqa
                          '</ul>\n'                                            # noqa
                        '</li>\n'                                              # noqa
                      '</ul>\n'                                                # noqa
                    '</div>\n'                                                 # noqa
                ),
                'toc_tokens': [
                    {
                        'level': 2,
                        'id': 'some-header',
                        'name': 'Some Header',
                        'html': 'Some Header',
                        'data-toc-label': '',
                        'children': [
                            {
                                'level': 3,
                                'id': 'next-level',
                                'name': 'Next Level',
                                'html': 'Next Level',
                                'data-toc-label': '',
                                'children': []
                            }
                        ]
                    }
                ]
            },
            extensions=[TocExtension(toc_depth=3, baselevel=2)]
        )

    def test_escaped_code(self):
        self.assertMarkdownRenders(
            self.dedent(
                '''
                [TOC]

                # `<test>`
                '''
            ),
            self.dedent(
                '''
                <div class="toc">
                <ul>
                <li><a href="#test">&lt;test&gt;</a></li>
                </ul>
                </div>
                <h1 id="test"><code>&lt;test&gt;</code></h1>
                '''
            ),
            extensions=['toc']
        )

    def test_escaped_char_in_id(self):
        self.assertMarkdownRenders(
            r'# escaped\_character',
            '<h1 id="escaped_character">escaped_character</h1>',
            expected_attrs={
                'toc': (
                    '<div class="toc">\n'
                      '<ul>\n'                                                           # noqa
                        '<li><a href="#escaped_character">escaped_character</a></li>\n'  # noqa
                      '</ul>\n'                                                          # noqa
                    '</div>\n'                                                           # noqa
                ),
                'toc_tokens': [
                    {
                        'level': 1,
                        'id': 'escaped_character',
                        'name': 'escaped_character',
                        'html': 'escaped_character',
                        'data-toc-label': '',
                        'children': []
                    }
                ]
            },
            extensions=['toc']
        )

    def test_escaped_char_in_attr_list(self):
        self.assertMarkdownRenders(
            r'# `*Foo*` { id="\*Foo\*" }',
            '<h1 id="*Foo*"><code>*Foo*</code></h1>',
            expected_attrs={
                'toc': (
                    '<div class="toc">\n'
                      '<ul>\n'                                                           # noqa
                        '<li><a href="#*Foo*">*Foo*</a></li>\n'                          # noqa
                      '</ul>\n'                                                          # noqa
                    '</div>\n'                                                           # noqa
                ),
                'toc_tokens': [
                    {
                        'level': 1,
                        'id': '*Foo*',
                        'name': '*Foo*',
                        'html': '<code>*Foo*</code>',
                        'data-toc-label': '',
                        'children': []
                    }
                ]
            },
            extensions=['toc', 'attr_list']
        )

    def testAutoLinkEmail(self):
        self.assertMarkdownRenders(
            '## <foo@example.org>',
            '<h2 id="fooexampleorg"><a href="&#109;&#97;&#105;&#108;&#116;&#111;&#58;&#102;&#111;&#111;&#64;&#101;'
            '&#120;&#97;&#109;&#112;&#108;&#101;&#46;&#111;&#114;&#103;">&#102;&#111;&#111;&#64;&#101;&#120;&#97;'
            '&#109;&#112;&#108;&#101;&#46;&#111;&#114;&#103;</a></h2>',
            expected_attrs={
                'toc_tokens': [
                    {
                        'level': 2,
                        'id': 'fooexampleorg',
                        'name': '&#102;&#111;&#111;&#64;&#101;&#120;&#97;&#109;'
                                '&#112;&#108;&#101;&#46;&#111;&#114;&#103;',
                        'html': '<a href="&#109;&#97;&#105;&#108;&#116;&#111;&#58;&#102;&#111;&#111;&#64;&#101;'
                                '&#120;&#97;&#109;&#112;&#108;&#101;&#46;&#111;&#114;&#103;">&#102;&#111;&#111;'
                                '&#64;&#101;&#120;&#97;&#109;&#112;&#108;&#101;&#46;&#111;&#114;&#103;</a>',
                        'data-toc-label': '',
                        'children': []
                    }
                ]
            },
            extensions=['toc']
        )

    def testAnchorLinkWithCustomClass(self):
        self.assertMarkdownRenders(
            self.dedent(
                '''
                # Header 1

                ## Header *2*
                '''
            ),
            self.dedent(
                '''
                <h1 id="header-1"><a class="custom" href="#header-1">Header 1</a></h1>
                <h2 id="header-2"><a class="custom" href="#header-2">Header <em>2</em></a></h2>
                '''
            ),
            extensions=[TocExtension(anchorlink=True, anchorlink_class="custom")]
        )

    def testAnchorLinkWithCustomClasses(self):
        self.assertMarkdownRenders(
            self.dedent(
                '''
                # Header 1

                ## Header *2*
                '''
            ),
            self.dedent(
                '''
                <h1 id="header-1"><a class="custom1 custom2" href="#header-1">Header 1</a></h1>
                <h2 id="header-2"><a class="custom1 custom2" href="#header-2">Header <em>2</em></a></h2>
                '''
            ),
            extensions=[TocExtension(anchorlink=True, anchorlink_class="custom1 custom2")]
        )

    def testPermalinkWithEmptyText(self):
        self.assertMarkdownRenders(
            '# Header',
            '<h1 id="header">'                                                      # noqa
                'Header'                                                            # noqa
                '<a class="headerlink" href="#header" title="Permanent link"></a>'  # noqa
            '</h1>',                                                                # noqa
            extensions=[TocExtension(permalink="")]
        )

    def testPermalinkWithCustomClass(self):
        self.assertMarkdownRenders(
            '# Header',
            '<h1 id="header">'                                                        # noqa
                'Header'                                                              # noqa
                '<a class="custom" href="#header" title="Permanent link">&para;</a>'  # noqa
            '</h1>',                                                                  # noqa
            extensions=[TocExtension(permalink=True, permalink_class="custom")]
        )

    def testPermalinkWithCustomClasses(self):
        self.assertMarkdownRenders(
            '# Header',
            '<h1 id="header">'                                                                 # noqa
                'Header'                                                                       # noqa
                '<a class="custom1 custom2" href="#header" title="Permanent link">&para;</a>'  # noqa
            '</h1>',                                                                           # noqa
            extensions=[TocExtension(permalink=True, permalink_class="custom1 custom2")]
        )

    def testPermalinkWithCustomTitle(self):
        self.assertMarkdownRenders(
            '# Header',
            '<h1 id="header">'                                                    # noqa
                'Header'                                                          # noqa
                '<a class="headerlink" href="#header" title="custom">&para;</a>'  # noqa
            '</h1>',                                                              # noqa
            extensions=[TocExtension(permalink=True, permalink_title="custom")]
        )

    def testPermalinkWithEmptyTitle(self):
        self.assertMarkdownRenders(
            '# Header',
            '<h1 id="header">'                                                    # noqa
                'Header'                                                          # noqa
                '<a class="headerlink" href="#header">&para;</a>'                 # noqa
            '</h1>',                                                              # noqa
            extensions=[TocExtension(permalink=True, permalink_title="")]
        )

    def testPermalinkWithUnicodeInID(self):
        from markdown.extensions.toc import slugify_unicode
        self.assertMarkdownRenders(
            '# Unicode ヘッダー',
            '<h1 id="unicode-ヘッダー">'                                                            # noqa
                'Unicode ヘッダー'                                                                  # noqa
                '<a class="headerlink" href="#unicode-ヘッダー" title="Permanent link">&para;</a>'  # noqa
            '</h1>',                                                                               # noqa
            extensions=[TocExtension(permalink=True, slugify=slugify_unicode)]
        )

    def testPermalinkWithUnicodeTitle(self):
        from markdown.extensions.toc import slugify_unicode
        self.assertMarkdownRenders(
            '# Unicode ヘッダー',
            '<h1 id="unicode-ヘッダー">'                                                        # noqa
                'Unicode ヘッダー'                                                              # noqa
                '<a class="headerlink" href="#unicode-ヘッダー" title="パーマリンク">&para;</a>'  # noqa
            '</h1>',                                                                           # noqa
            extensions=[TocExtension(permalink=True, permalink_title="パーマリンク", slugify=slugify_unicode)]
        )

    def testPermalinkWithExtendedLatinInID(self):
        self.assertMarkdownRenders(
            '# Théâtre',
            '<h1 id="theatre">'                                                            # noqa
                'Théâtre'                                                                  # noqa
                '<a class="headerlink" href="#theatre" title="Permanent link">&para;</a>'  # noqa
            '</h1>',                                                                       # noqa
            extensions=[TocExtension(permalink=True)]
        )

    def testNl2brCompatibility(self):
        self.assertMarkdownRenders(
            '[TOC]\ntext',
            '<p>[TOC]<br />\ntext</p>',
            extensions=[TocExtension(), Nl2BrExtension()]
        )

    def testTOCWithCustomClass(self):

        self.assertMarkdownRenders(
            self.dedent(
                '''
                [TOC]
                # Header
                '''
            ),
            self.dedent(
                '''
                <div class="custom">
                <ul>
                <li><a href="#header">Header</a></li>
                </ul>
                </div>
                <h1 id="header">Header</h1>
                '''
            ),
            extensions=[TocExtension(toc_class="custom")]
        )

    def testTOCWithCustomClasses(self):
        self.assertMarkdownRenders(
            self.dedent(
                '''
                [TOC]
                # Header
                '''
            ),
            self.dedent(
                '''
                <div class="custom1 custom2">
                <ul>
                <li><a href="#header">Header</a></li>
                </ul>
                </div>
                <h1 id="header">Header</h1>
                '''
            ),
            extensions=[TocExtension(toc_class="custom1 custom2")]
        )

    def testTOCWithEmptyTitleClass(self):

        self.assertMarkdownRenders(
            self.dedent(
                '''
                [TOC]
                # Header
                '''
            ),
            self.dedent(
                '''
                <div class="toc"><span>ToC</span><ul>
                <li><a href="#header">Header</a></li>
                </ul>
                </div>
                <h1 id="header">Header</h1>
                '''
            ),
            extensions=[TocExtension(title_class="", title='ToC')]
        )

    def testTOCWithCustomTitleClass(self):

        self.assertMarkdownRenders(
            self.dedent(
                '''
                [TOC]
                # Header
                '''
            ),
            self.dedent(
                '''
                <div class="toc"><span class="tocname">ToC</span><ul>
                <li><a href="#header">Header</a></li>
                </ul>
                </div>
                <h1 id="header">Header</h1>
                '''
            ),
            extensions=[TocExtension(title_class="tocname", title='ToC')]
        )

    def testTocWithAttrList(self):

        self.assertMarkdownRenders(
            self.dedent(
                '''
                # Header 1

                ## Header 2 { #foo }

                ## Header 3 { data-toc-label="Foo Bar" }

                # Header 4 { data-toc-label="Foo > &amp; < Baz" }

                # Header 5 { data-toc-label="Foo <b>Quux</b>" }
                '''
            ),
            self.dedent(
                '''
                <h1 id="header-1">Header 1</h1>
                <h2 id="foo">Header 2</h2>
                <h2 id="header-3">Header 3</h2>
                <h1 id="header-4">Header 4</h1>
                <h1 id="header-5">Header 5</h1>
                '''
            ),
            expected_attrs={
                'toc': (
                    '<div class="toc">\n'
                      '<ul>\n'                                                        # noqa
                        '<li><a href="#header-1">Header 1</a>'                        # noqa
                          '<ul>\n'                                                    # noqa
                            '<li><a href="#foo">Header 2</a></li>\n'                  # noqa
                            '<li><a href="#header-3">Foo Bar</a></li>\n'              # noqa
                          '</ul>\n'                                                   # noqa
                        '</li>\n'                                                     # noqa
                        '<li><a href="#header-4">Foo &gt; &amp; &lt; Baz</a></li>\n'  # noqa
                        '<li><a href="#header-5">Foo Quux</a></li>\n'                 # noqa
                      '</ul>\n'                                                       # noqa
                    '</div>\n'
                ),
                'toc_tokens': [
                    {
                        'level': 1,
                        'id': 'header-1',
                        'name': 'Header 1',
                        'html': 'Header 1',
                        'data-toc-label': '',
                        'children': [
                            {
                                'level': 2,
                                'id': 'foo',
                                'name': 'Header 2',
                                'html': 'Header 2',
                                'data-toc-label': '',
                                'children': []
                            },
                            {
                                'level': 2,
                                'id': 'header-3',
                                'name': 'Foo Bar',
                                'html': 'Header 3',
                                'data-toc-label': 'Foo Bar',
                                'children': []
                            }
                        ]
                    },
                    {
                        'level': 1,
                        'id': 'header-4',
                        'name': 'Foo &gt; &amp; &lt; Baz',
                        'html': 'Header 4',
                        'data-toc-label': 'Foo > &amp; < Baz',
                        'children': []
                    },
                    {
                        'level': 1,
                        'id': 'header-5',
                        'name': 'Foo Quux',
                        'html': 'Header 5',
                        'data-toc-label': 'Foo <b>Quux</b>',
                        'children': []
                    },
                ]
            },
            extensions=[TocExtension(), 'attr_list']
        )

    def testHeadingRemoveFootnoteRef(self):

        self.assertMarkdownRenders(
            self.dedent(
                '''
                # Header 1[^1]
                # Header[^1] 2
                # Header *subelement*[^1] 3
                # Header[^1] double[^1] 4

                [^1]: footnote
                '''
            ),
            (
                '<h1 id="header-1">Header 1<sup id="fnref:1"><a class="footnote-ref" href="#fn:1">1</a></sup></h1>\n'
                '<h1 id="header-2">Header<sup id="fnref2:1"><a class="footnote-ref" href="#fn:1">1</a></sup> 2</h1>\n'
                '<h1 id="header-subelement-3">'
                'Header <em>subelement</em><sup id="fnref3:1"><a class="footnote-ref" href="#fn:1">1</a></sup> 3'
                '</h1>\n'
                '<h1 id="header-double-4">'
                'Header<sup id="fnref4:1"><a class="footnote-ref" href="#fn:1">1</a></sup> double'
                '<sup id="fnref5:1"><a class="footnote-ref" href="#fn:1">1</a></sup> 4'
                '</h1>\n'
                '<div class="footnote">\n'
                '<hr />\n'
                '<ol>\n'
                '<li id="fn:1">\n'
                '<p>'
                'footnote&#160;'
                '<a class="footnote-backref" href="#fnref:1" title="Jump back to footnote 1 in the text">&#8617;</a>'
                '<a class="footnote-backref" href="#fnref2:1" title="Jump back to footnote 1 in the text">&#8617;</a>'
                '<a class="footnote-backref" href="#fnref3:1" title="Jump back to footnote 1 in the text">&#8617;</a>'
                '<a class="footnote-backref" href="#fnref4:1" title="Jump back to footnote 1 in the text">&#8617;</a>'
                '<a class="footnote-backref" href="#fnref5:1" title="Jump back to footnote 1 in the text">&#8617;</a>'
                '</p>\n'
                '</li>\n'
                '</ol>\n'
                '</div>'
            ),
            expected_attrs={
                'toc': (
                    '<div class="toc">\n'
                      '<ul>\n'                                                               # noqa
                        '<li><a href="#header-1">Header 1</a></li>\n'                        # noqa
                        '<li><a href="#header-2">Header 2</a></li>\n'                        # noqa
                        '<li><a href="#header-subelement-3">Header subelement 3</a></li>\n'  # noqa
                        '<li><a href="#header-double-4">Header double 4</a></li>\n'          # noqa
                      '</ul>\n'                                                              # noqa
                    '</div>\n'                                                               # noqa
                ),
                'toc_tokens': [
                    {
                        'level': 1,
                        'id': 'header-1',
                        'name': 'Header 1',
                        'html': 'Header 1',
                        'data-toc-label': '',
                        'children': []
                    },
                    {
                        'level': 1,
                        'id': 'header-2',
                        'name': 'Header 2',
                        'html': 'Header 2',
                        'data-toc-label': '',
                        'children': []
                    },
                    {
                        'level': 1,
                        'id': 'header-subelement-3',
                        'name': 'Header subelement 3',
                        'html': 'Header <em>subelement</em> 3',
                        'data-toc-label': '',
                        'children': []
                    },
                    {
                        'level': 1,
                        'id': 'header-double-4',
                        'name': 'Header double 4',
                        'html': 'Header double 4',
                        'data-toc-label': '',
                        'children': []
                    }
                ]
            },
            extensions=[TocExtension(), 'footnotes']
        )


class testStripTags(TestCase):

    def testStripElement(self):
        self.assertEqual(
            strip_tags('foo <em>bar</em>'),
            'foo bar'
        )

    def testStripOpenElement(self):
        self.assertEqual(
            strip_tags('foo <em>bar'),
            'foo bar'
        )

    def testStripEmptyElement(self):
        self.assertEqual(
            strip_tags('foo <br />bar'),
            'foo bar'
        )

    def testDontStripOpenBracket(self):
        self.assertEqual(
            strip_tags('foo < bar'),
            'foo < bar'
        )

    def testDontStripCloseBracket(self):
        self.assertEqual(
            strip_tags('foo > bar'),
            'foo > bar'
        )

    def testStripCollapseWhitespace(self):
        self.assertEqual(
            strip_tags('foo <em>\tbar\t</em>'),
            'foo bar'
        )

    def testStripElementWithNewlines(self):
        self.assertEqual(
            strip_tags('foo <meta content="tag\nwith\nnewlines"> bar'),
            'foo bar'
        )

    def testStripComment(self):
        self.assertEqual(
            strip_tags('foo <!-- comment --> bar'),
            'foo bar'
        )

    def testStripCommentWithInnerTags(self):
        self.assertEqual(
            strip_tags('foo <!-- comment with <em> --> bar'),
            'foo bar'
        )

    def testStripCommentInElement(self):
        self.assertEqual(
            strip_tags('<em>foo <!-- comment --> bar<em>'),
            'foo bar'
        )

    def testDontStripHTMLEntities(self):
        self.assertEqual(
            strip_tags('foo &lt; &amp; &lt; bar'),
            'foo &lt; &amp; &lt; bar'
        )



"""
Python Markdown

A Python implementation of John Gruber's Markdown.

Documentation: https://python-markdown.github.io/
GitHub: https://github.com/Python-Markdown/markdown/
PyPI: https://pypi.org/project/Markdown/

Started by Manfred Stienstra (http://www.dwerg.net/).
Maintained for a few years by Yuri Takhteyev (http://www.freewisdom.org).
Currently maintained by Waylan Limberg (https://github.com/waylan),
Dmitry Shachnev (https://github.com/mitya57) and Isaac Muse (https://github.com/facelessuser).

Copyright 2007-2019 The Python Markdown Project (v. 1.7 and later)
Copyright 2004, 2005, 2006 Yuri Takhteyev (v. 0.2-1.6b)
Copyright 2004 Manfred Stienstra (the original version)

License: BSD (see LICENSE.md for details).
"""

from markdown.test_tools import TestCase


class TestDefList(TestCase):

    def test_def_list_with_ol(self):
        self.assertMarkdownRenders(
            self.dedent(
                '''

                term

                :   this is a definition for term. it has
                    multiple lines in the first paragraph.

                    1.  first thing

                        first thing details in a second paragraph.

                    1.  second thing

                        second thing details in a second paragraph.

                    1.  third thing

                        third thing details in a second paragraph.
                '''
            ),
            self.dedent(
                '''
                <dl>
                <dt>term</dt>
                <dd>
                <p>this is a definition for term. it has
                multiple lines in the first paragraph.</p>
                <ol>
                <li>
                <p>first thing</p>
                <p>first thing details in a second paragraph.</p>
                </li>
                <li>
                <p>second thing</p>
                <p>second thing details in a second paragraph.</p>
                </li>
                <li>
                <p>third thing</p>
                <p>third thing details in a second paragraph.</p>
                </li>
                </ol>
                </dd>
                </dl>
                '''
            ),
            extensions=['def_list']
        )

    def test_def_list_with_ul(self):
        self.assertMarkdownRenders(
            self.dedent(
                '''

                term

                :   this is a definition for term. it has
                    multiple lines in the first paragraph.

                    -   first thing

                        first thing details in a second paragraph.

                    -   second thing

                        second thing details in a second paragraph.

                    -   third thing

                        third thing details in a second paragraph.
                '''
            ),
            self.dedent(
                '''
                <dl>
                <dt>term</dt>
                <dd>
                <p>this is a definition for term. it has
                multiple lines in the first paragraph.</p>
                <ul>
                <li>
                <p>first thing</p>
                <p>first thing details in a second paragraph.</p>
                </li>
                <li>
                <p>second thing</p>
                <p>second thing details in a second paragraph.</p>
                </li>
                <li>
                <p>third thing</p>
                <p>third thing details in a second paragraph.</p>
                </li>
                </ul>
                </dd>
                </dl>
                '''
            ),
            extensions=['def_list']
        )

    def test_def_list_with_nesting(self):
        self.assertMarkdownRenders(
            self.dedent(
                '''

                term

                :   this is a definition for term. it has
                    multiple lines in the first paragraph.

                    1.  first thing

                        first thing details in a second paragraph.

                        -   first nested thing

                            second nested thing details
                '''
            ),
            self.dedent(
                '''
                <dl>
                <dt>term</dt>
                <dd>
                <p>this is a definition for term. it has
                multiple lines in the first paragraph.</p>
                <ol>
                <li>
                <p>first thing</p>
                <p>first thing details in a second paragraph.</p>
                <ul>
                <li>
                <p>first nested thing</p>
                <p>second nested thing details</p>
                </li>
                </ul>
                </li>
                </ol>
                </dd>
                </dl>
                '''
            ),
            extensions=['def_list']
        )

    def test_def_list_with_nesting_self(self):
        self.assertMarkdownRenders(
            self.dedent(
                '''

                term

                :   this is a definition for term. it has
                    multiple lines in the first paragraph.

                    inception

                    :   this is a definition for term. it has
                        multiple lines in the first paragraph.

                        - bullet point

                          another paragraph
                '''
            ),
            self.dedent(
                '''
                <dl>
                <dt>term</dt>
                <dd>
                <p>this is a definition for term. it has
                multiple lines in the first paragraph.</p>
                <dl>
                <dt>inception</dt>
                <dd>
                <p>this is a definition for term. it has
                multiple lines in the first paragraph.</p>
                <ul>
                <li>bullet point</li>
                </ul>
                <p>another paragraph</p>
                </dd>
                </dl>
                </dd>
                </dl>
                '''
            ),
            extensions=['def_list']
        )

    def test_def_list_unreasonable_nesting(self):
        self.assertMarkdownRenders(
            self.dedent(
                '''

                turducken

                :   this is a definition for term. it has
                    multiple lines in the first paragraph.

                    1.  ordered list

                        - nested list

                            term

                            :   definition

                                -   item 1 paragraph 1

                                    item 1 paragraph 2
                '''
            ),
            self.dedent(
                '''
                <dl>
                <dt>turducken</dt>
                <dd>
                <p>this is a definition for term. it has
                multiple lines in the first paragraph.</p>
                <ol>
                <li>
                <p>ordered list</p>
                <ul>
                <li>
                <p>nested list</p>
                <dl>
                <dt>term</dt>
                <dd>
                <p>definition</p>
                <ul>
                <li>
                <p>item 1 paragraph 1</p>
                <p>item 1 paragraph 2</p>
                </li>
                </ul>
                </dd>
                </dl>
                </li>
                </ul>
                </li>
                </ol>
                </dd>
                </dl>
                '''
            ),
            extensions=['def_list']
        )

    def test_def_list_nested_admontions(self):
        self.assertMarkdownRenders(
            self.dedent(
                '''
                term

                :   definition

                    !!! note "Admontion"

                        term

                        :   definition

                            1.  list

                                continue
                '''
            ),
            self.dedent(
                '''
                <dl>
                <dt>term</dt>
                <dd>
                <p>definition</p>
                <div class="admonition note">
                <p class="admonition-title">Admontion</p>
                <dl>
                <dt>term</dt>
                <dd>
                <p>definition</p>
                <ol>
                <li>
                <p>list</p>
                <p>continue</p>
                </li>
                </ol>
                </dd>
                </dl>
                </div>
                </dd>
                </dl>
                '''
            ),
            extensions=['def_list', 'admonition']
        )



"""Griffe extensions."""

from __future__ import annotations

import ast
from typing import TYPE_CHECKING, Any
import textwrap

from griffe import Docstring, Extension, DocstringSectionAdmonition, DocstringSectionText, Visitor, Inspector

if TYPE_CHECKING:
    from griffe import Class, Function, ObjectNode


def _deprecated(obj: Class | Function) -> str | None:
    for decorator in obj.decorators:
        if decorator.callable_path == "markdown.util.deprecated":
            return ast.literal_eval(str(decorator.value.arguments[0]))
    return None


class DeprecatedExtension(Extension):
    """Griffe extension for `@markdown.util.deprecated` decorator support."""

    def _insert_message(self, obj: Function | Class, message: str) -> None:
        if not obj.docstring:
            obj.docstring = Docstring("", parent=obj)
        sections = obj.docstring.parsed
        sections.insert(0, DocstringSectionAdmonition(kind="warning", text=message, title="Deprecated"))

    def on_class_instance(self, node: ast.AST | ObjectNode, cls: Class, agent: Visitor | Inspector, **kwargs: Any) -> None:  # noqa: ARG002
        """Add section to docstrings of deprecated classes."""
        if message := _deprecated(cls):
            self._insert_message(cls, message)
            cls.labels.add("deprecated")

    def on_function_instance(self, node: ast.AST | ObjectNode, func: Function, agent: Visitor | Inspector, **kwargs: Any) -> None:  # noqa: ARG002
        """Add section to docstrings of deprecated functions."""
        if message := _deprecated(func):
            self._insert_message(func, message)
            func.labels.add("deprecated")


class PriorityTableExtension(Extension):
    """ Griffe extension to insert a table of processor priority in specified functions. """

    def __init__(self, paths: list[str] | None = None) -> None:
        super().__init__()
        self.paths = paths

    def linked_obj(self, value: str, path: str) -> str:
        """ Wrap object name in reference link. """
        return f'[`{value}`][{path}.{value}]'

    def on_function_instance(self, node: ast.AST | ObjectNode, func: Function, agent: Visitor | Inspector, **kwargs: Any) -> None:  # noqa: ARG002
        """Add table to specified function docstrings."""
        if self.paths and func.path not in self.paths:
            return  # skip objects that were not selected

        # Table header
        data = [
            'Class Instance | Name | Priority',
            '-------------- | ---- | :------:'
        ]

        # Extract table body from source code of function.
        for obj in node.body:
            # Extract the arguments passed to `util.Registry.register`.
            if isinstance(obj, ast.Expr) and isinstance(obj.value, ast.Call) and obj.value.func.attr == 'register':
                _args = obj.value.args
                cls = self.linked_obj(_args[0].func.id, func.path.rsplit('.', 1)[0])
                name = _args[1].value
                priority = str(_args[2].value)
                if func.name == ('build_inlinepatterns'):
                    # Include Pattern: first arg passed to class
                    if isinstance(_args[0].args[0], ast.Constant):
                        # Pattern is a string
                        value = f'`"{_args[0].args[0].value}"`'
                    else:
                        # Pattern is a variable
                        value = self.linked_obj(_args[0].args[0].id, func.path.rsplit('.', 1)[0])
                    cls = f'{cls}({value})'
                data.append(f'{cls} | `{name}` | `{priority}`')

        table = '\n'.join(data)
        body = (
            f"Return a [`{func.returns.canonical_name}`][{func.returns.canonical_path}] instance which contains "
            "the following collection of classes with their assigned names and priorities.\n\n"
            f"{table}"
        )

        # Add to docstring.
        if not func.docstring:
            func.docstring = Docstring("", parent=func)
        sections = func.docstring.parsed
        sections.append(DocstringSectionText(body, title="Priority Table"))



"""
Python Markdown

A Python implementation of John Gruber's Markdown.

Documentation: https://python-markdown.github.io/
GitHub: https://github.com/Python-Markdown/markdown/
PyPI: https://pypi.org/project/Markdown/

Started by Manfred Stienstra (http://www.dwerg.net/).
Maintained for a few years by Yuri Takhteyev (http://www.freewisdom.org).
Currently maintained by Waylan Limberg (https://github.com/waylan),
Dmitry Shachnev (https://github.com/mitya57) and Isaac Muse (https://github.com/facelessuser).

Copyright 2007-2024 The Python Markdown Project (v. 1.7 and later)
Copyright 2004, 2005, 2006 Yuri Takhteyev (v. 0.2-1.6b)
Copyright 2004 Manfred Stienstra (the original version)

License: BSD (see LICENSE.md for details).
"""

import unittest
from markdown.test_tools import TestCase


class TestUnorderedLists(TestCase):

    # TODO: Move legacy tests here

    def test_header_and_paragraph_no_blank_line_loose_list(self):
        self.assertMarkdownRenders(
            self.dedent(
                """
                - ### List 1
                    Entry 1.1

                - ### List 2
                    Entry 2.1
                """
            ),
            self.dedent(
                """
                <ul>
                <li>
                <h3>List 1</h3>
                <p>Entry 1.1</p>
                </li>
                <li>
                <h3>List 2</h3>
                <p>Entry 2.1</p>
                </li>
                </ul>
                """
            )
        )

    # Note: This is strange. Any other element on first line of list item would get very
    # different behavior. However, as a heading can only ever be one line, this is the
    # correct behavior. In fact, `markdown.pl` behaves this way with no indentation.
    def test_header_and_paragraph_no_blank_line_loose_list_no_indent(self):
        self.assertMarkdownRenders(
            self.dedent(
                """
                - ### List 1
                Entry 1.1

                - ### List 2
                Entry 2.1
                """
            ),
            self.dedent(
                """
                <ul>
                <li>
                <h3>List 1</h3>
                <p>Entry 1.1</p>
                </li>
                <li>
                <h3>List 2</h3>
                <p>Entry 2.1</p>
                </li>
                </ul>
                """
            )
        )

    # TODO: Possibly change this behavior. While this test follows the behavior
    # of `markdown.pl`, it is likely surprising to most users. In fact, actual
    # behavior is to return the same results as for a loose list.
    @unittest.skip('This behaves as a loose list in Python-Markdown')
    def test_header_and_paragraph_no_blank_line_tight_list(self):
        self.assertMarkdownRenders(
            self.dedent(
                """
                - ### List 1
                    Entry 1.1
                - ### List 2
                    Entry 2.1
                """
            ),
            self.dedent(
                """
                <ul>
                <li>### List 1
                Entry 1.1</li>
                <li>### List 2
                Entry 2.1</li>
                </ul>
                """
            )
        )

    # TODO: Possibly change this behavior. While this test follows the behavior
    # of `markdown.pl`, it is likely surprising to most users. In fact, actual
    # behavior is to return the same results as for a loose list.
    @unittest.skip('This behaves as a loose list in Python-Markdown')
    def test_header_and_paragraph_no_blank_line_tight_list_no_indent(self):
        self.assertMarkdownRenders(
            self.dedent(
                """
                - ### List 1
                Entry 1.1
                - ### List 2
                Entry 2.1
                """
            ),
            self.dedent(
                """
                <ul>
                <li>### List 1
                Entry 1.1</li>
                <li>### List 2
                Entry 2.1</li>
                </ul>
                """
            )
        )



"""
Python Markdown

A Python implementation of John Gruber's Markdown.

Documentation: https://python-markdown.github.io/
GitHub: https://github.com/Python-Markdown/markdown/
PyPI: https://pypi.org/project/Markdown/

Started by Manfred Stienstra (http://www.dwerg.net/).
Maintained for a few years by Yuri Takhteyev (http://www.freewisdom.org).
Currently maintained by Waylan Limberg (https://github.com/waylan),
Dmitry Shachnev (https://github.com/mitya57) and Isaac Muse (https://github.com/facelessuser).

Copyright 2007-2019 The Python Markdown Project (v. 1.7 and later)
Copyright 2004, 2005, 2006 Yuri Takhteyev (v. 0.2-1.6b)
Copyright 2004 Manfred Stienstra (the original version)

License: BSD (see LICENSE.md for details).
"""

from markdown.test_tools import TestCase


class TestAdmonition(TestCase):

    def test_with_lists(self):
        self.assertMarkdownRenders(
            self.dedent(
                '''
                - List

                    !!! note "Admontion"

                        - Paragraph

                            Paragraph
                '''
            ),
            self.dedent(
                '''
                <ul>
                <li>
                <p>List</p>
                <div class="admonition note">
                <p class="admonition-title">Admontion</p>
                <ul>
                <li>
                <p>Paragraph</p>
                <p>Paragraph</p>
                </li>
                </ul>
                </div>
                </li>
                </ul>
                '''
            ),
            extensions=['admonition']
        )

    def test_with_big_lists(self):
        self.assertMarkdownRenders(
            self.dedent(
                '''
                - List

                    !!! note "Admontion"

                        - Paragraph

                            Paragraph

                        - Paragraph

                            paragraph
                '''
            ),
            self.dedent(
                '''
                <ul>
                <li>
                <p>List</p>
                <div class="admonition note">
                <p class="admonition-title">Admontion</p>
                <ul>
                <li>
                <p>Paragraph</p>
                <p>Paragraph</p>
                </li>
                <li>
                <p>Paragraph</p>
                <p>paragraph</p>
                </li>
                </ul>
                </div>
                </li>
                </ul>
                '''
            ),
            extensions=['admonition']
        )

    def test_with_complex_lists(self):
        self.assertMarkdownRenders(
            self.dedent(
                '''
                - List

                    !!! note "Admontion"

                        - Paragraph

                            !!! note "Admontion"

                                1. Paragraph

                                    Paragraph
                '''
            ),
            self.dedent(
                '''
                <ul>
                <li>
                <p>List</p>
                <div class="admonition note">
                <p class="admonition-title">Admontion</p>
                <ul>
                <li>
                <p>Paragraph</p>
                <div class="admonition note">
                <p class="admonition-title">Admontion</p>
                <ol>
                <li>
                <p>Paragraph</p>
                <p>Paragraph</p>
                </li>
                </ol>
                </div>
                </li>
                </ul>
                </div>
                </li>
                </ul>
                '''
            ),
            extensions=['admonition']
        )

    def test_definition_list(self):
        self.assertMarkdownRenders(
            self.dedent(
                '''
                - List

                    !!! note "Admontion"

                        Term

                        :   Definition

                            More text

                        :   Another
                            definition

                            Even more text
                '''
            ),
            self.dedent(
                '''
                <ul>
                <li>
                <p>List</p>
                <div class="admonition note">
                <p class="admonition-title">Admontion</p>
                <dl>
                <dt>Term</dt>
                <dd>
                <p>Definition</p>
                <p>More text</p>
                </dd>
                <dd>
                <p>Another
                definition</p>
                <p>Even more text</p>
                </dd>
                </dl>
                </div>
                </li>
                </ul>
                '''
            ),
            extensions=['admonition', 'def_list']
        )

    def test_with_preceding_text(self):
        self.assertMarkdownRenders(
            self.dedent(
                '''
                foo
                **foo**
                !!! note "Admonition"
                '''
            ),
            self.dedent(
                '''
                <p>foo
                <strong>foo</strong></p>
                <div class="admonition note">
                <p class="admonition-title">Admonition</p>
                </div>
                '''
            ),
            extensions=['admonition']
        )

    def test_admontion_detabbing(self):
        self.assertMarkdownRenders(
            self.dedent(
                '''
                !!! note "Admonition"
                    - Parent 1

                        - Child 1
                        - Child 2
                '''
            ),
            self.dedent(
                '''
                <div class="admonition note">
                <p class="admonition-title">Admonition</p>
                <ul>
                <li>
                <p>Parent 1</p>
                <ul>
                <li>Child 1</li>
                <li>Child 2</li>
                </ul>
                </li>
                </ul>
                </div>
                '''
            ),
            extensions=['admonition']
        )

    def test_admonition_first_indented(self):
        self.assertMarkdownRenders(
            self.dedent(
                '''
                !!! danger "This is not"
                        one long admonition title
                '''
            ),
            self.dedent(
                '''
                <div class="admonition danger">
                <p class="admonition-title">This is not</p>
                <pre><code>one long admonition title
                </code></pre>
                </div>
                '''
            ),
            extensions=['admonition']
        )



"""
Python Markdown

A Python implementation of John Gruber's Markdown.

Documentation: https://python-markdown.github.io/
GitHub: https://github.com/Python-Markdown/markdown/
PyPI: https://pypi.org/project/Markdown/

Started by Manfred Stienstra (http://www.dwerg.net/).
Maintained for a few years by Yuri Takhteyev (http://www.freewisdom.org).
Currently maintained by Waylan Limberg (https://github.com/waylan),
Dmitry Shachnev (https://github.com/mitya57) and Isaac Muse (https://github.com/facelessuser).

Copyright 2007-2023 The Python Markdown Project (v. 1.7 and later)
Copyright 2004, 2005, 2006 Yuri Takhteyev (v. 0.2-1.6b)
Copyright 2004 Manfred Stienstra (the original version)

License: BSD (see LICENSE.md for details).
"""

import unittest
from markdown.test_tools import TestCase


class TestSetextHeaders(TestCase):

    def test_setext_h1(self):
        self.assertMarkdownRenders(
            self.dedent(
                """
                This is an H1
                =============
                """
            ),

            '<h1>This is an H1</h1>'
        )

    def test_setext_h2(self):
        self.assertMarkdownRenders(
            self.dedent(
                """
                This is an H2
                -------------
                """
            ),

            '<h2>This is an H2</h2>'
        )

    def test_setext_h1_mismatched_length(self):
        self.assertMarkdownRenders(
            self.dedent(
                """
                This is an H1
                ===
                """
            ),

            '<h1>This is an H1</h1>'
        )

    def test_setext_h2_mismatched_length(self):
        self.assertMarkdownRenders(
            self.dedent(
                """
                This is an H2
                ---
                """
            ),

            '<h2>This is an H2</h2>'
        )

    def test_setext_h1_followed_by_p(self):
        self.assertMarkdownRenders(
            self.dedent(
                """
                This is an H1
                =============
                Followed by a Paragraph with no blank line.
                """
            ),
            self.dedent(
                """
                <h1>This is an H1</h1>
                <p>Followed by a Paragraph with no blank line.</p>
                """
            )
        )

    def test_setext_h2_followed_by_p(self):
        self.assertMarkdownRenders(
            self.dedent(
                """
                This is an H2
                -------------
                Followed by a Paragraph with no blank line.
                """
            ),
            self.dedent(
                """
                <h2>This is an H2</h2>
                <p>Followed by a Paragraph with no blank line.</p>
                """
            )
        )

    # TODO: fix this
    # see https://johnmacfarlane.net/babelmark2/?normalize=1&text=Paragraph%0AAn+H1%0A%3D%3D%3D%3D%3D
    @unittest.skip('This is broken in Python-Markdown')
    def test_p_followed_by_setext_h1(self):
        self.assertMarkdownRenders(
            self.dedent(
                """
                This is a Paragraph.
                Followed by an H1 with no blank line.
                =====================================
                """
            ),
            self.dedent(
                """
                <p>This is a Paragraph.</p>
                <h1>Followed by an H1 with no blank line.</h1>
                """
            )
        )

    # TODO: fix this
    # see https://johnmacfarlane.net/babelmark2/?normalize=1&text=Paragraph%0AAn+H2%0A-----
    @unittest.skip('This is broken in Python-Markdown')
    def test_p_followed_by_setext_h2(self):
        self.assertMarkdownRenders(
            self.dedent(
                """
                This is a Paragraph.
                Followed by an H2 with no blank line.
                -------------------------------------
                """
            ),
            self.dedent(
                """
                <p>This is a Paragraph.</p>
                <h2>Followed by an H2 with no blank line.</h2>
                """
            )
        )


class TestHashHeaders(TestCase):

    def test_hash_h1_open(self):
        self.assertMarkdownRenders(
            '# This is an H1',

            '<h1>This is an H1</h1>'
        )

    def test_hash_h2_open(self):
        self.assertMarkdownRenders(
            '## This is an H2',

            '<h2>This is an H2</h2>'
        )

    def test_hash_h3_open(self):
        self.assertMarkdownRenders(
            '### This is an H3',

            '<h3>This is an H3</h3>'
        )

    def test_hash_h4_open(self):
        self.assertMarkdownRenders(
            '#### This is an H4',

            '<h4>This is an H4</h4>'
        )

    def test_hash_h5_open(self):
        self.assertMarkdownRenders(
            '##### This is an H5',

            '<h5>This is an H5</h5>'
        )

    def test_hash_h6_open(self):
        self.assertMarkdownRenders(
            '###### This is an H6',

            '<h6>This is an H6</h6>'
        )

    def test_hash_gt6_open(self):
        self.assertMarkdownRenders(
            '####### This is an H6',

            '<h6># This is an H6</h6>'
        )

    def test_hash_h1_open_missing_space(self):
        self.assertMarkdownRenders(
            '#This is an H1',

            '<h1>This is an H1</h1>'
        )

    def test_hash_h2_open_missing_space(self):
        self.assertMarkdownRenders(
            '##This is an H2',

            '<h2>This is an H2</h2>'
        )

    def test_hash_h3_open_missing_space(self):
        self.assertMarkdownRenders(
            '###This is an H3',

            '<h3>This is an H3</h3>'
        )

    def test_hash_h4_open_missing_space(self):
        self.assertMarkdownRenders(
            '####This is an H4',

            '<h4>This is an H4</h4>'
        )

    def test_hash_h5_open_missing_space(self):
        self.assertMarkdownRenders(
            '#####This is an H5',

            '<h5>This is an H5</h5>'
        )

    def test_hash_h6_open_missing_space(self):
        self.assertMarkdownRenders(
            '######This is an H6',

            '<h6>This is an H6</h6>'
        )

    def test_hash_gt6_open_missing_space(self):
        self.assertMarkdownRenders(
            '#######This is an H6',

            '<h6>#This is an H6</h6>'
        )

    def test_hash_h1_closed(self):
        self.assertMarkdownRenders(
            '# This is an H1 #',

            '<h1>This is an H1</h1>'
        )

    def test_hash_h2_closed(self):
        self.assertMarkdownRenders(
            '## This is an H2 ##',

            '<h2>This is an H2</h2>'
        )

    def test_hash_h3_closed(self):
        self.assertMarkdownRenders(
            '### This is an H3 ###',

            '<h3>This is an H3</h3>'
        )

    def test_hash_h4_closed(self):
        self.assertMarkdownRenders(
            '#### This is an H4 ####',

            '<h4>This is an H4</h4>'
        )

    def test_hash_h5_closed(self):
        self.assertMarkdownRenders(
            '##### This is an H5 #####',

            '<h5>This is an H5</h5>'
        )

    def test_hash_h6_closed(self):
        self.assertMarkdownRenders(
            '###### This is an H6 ######',

            '<h6>This is an H6</h6>'
        )

    def test_hash_gt6_closed(self):
        self.assertMarkdownRenders(
            '####### This is an H6 #######',

            '<h6># This is an H6</h6>'
        )

    def test_hash_h1_closed_missing_space(self):
        self.assertMarkdownRenders(
            '#This is an H1#',

            '<h1>This is an H1</h1>'
        )

    def test_hash_h2_closed_missing_space(self):
        self.assertMarkdownRenders(
            '##This is an H2##',

            '<h2>This is an H2</h2>'
        )

    def test_hash_h3_closed_missing_space(self):
        self.assertMarkdownRenders(
            '###This is an H3###',

            '<h3>This is an H3</h3>'
        )

    def test_hash_h4_closed_missing_space(self):
        self.assertMarkdownRenders(
            '####This is an H4####',

            '<h4>This is an H4</h4>'
        )

    def test_hash_h5_closed_missing_space(self):
        self.assertMarkdownRenders(
            '#####This is an H5#####',

            '<h5>This is an H5</h5>'
        )

    def test_hash_h6_closed_missing_space(self):
        self.assertMarkdownRenders(
            '######This is an H6######',

            '<h6>This is an H6</h6>'
        )

    def test_hash_gt6_closed_missing_space(self):
        self.assertMarkdownRenders(
            '#######This is an H6#######',

            '<h6>#This is an H6</h6>'
        )

    def test_hash_h1_closed_mismatch(self):
        self.assertMarkdownRenders(
            '# This is an H1 ##',

            '<h1>This is an H1</h1>'
        )

    def test_hash_h2_closed_mismatch(self):
        self.assertMarkdownRenders(
            '## This is an H2 #',

            '<h2>This is an H2</h2>'
        )

    def test_hash_h3_closed_mismatch(self):
        self.assertMarkdownRenders(
            '### This is an H3 #',

            '<h3>This is an H3</h3>'
        )

    def test_hash_h4_closed_mismatch(self):
        self.assertMarkdownRenders(
            '#### This is an H4 #',

            '<h4>This is an H4</h4>'
        )

    def test_hash_h5_closed_mismatch(self):
        self.assertMarkdownRenders(
            '##### This is an H5 #',

            '<h5>This is an H5</h5>'
        )

    def test_hash_h6_closed_mismatch(self):
        self.assertMarkdownRenders(
            '###### This is an H6 #',

            '<h6>This is an H6</h6>'
        )

    def test_hash_gt6_closed_mismatch(self):
        self.assertMarkdownRenders(
            '####### This is an H6 ##################',

            '<h6># This is an H6</h6>'
        )

    def test_hash_h1_followed_by_p(self):
        self.assertMarkdownRenders(
            self.dedent(
                """
                # This is an H1
                Followed by a Paragraph with no blank line.
                """
            ),
            self.dedent(
                """
                <h1>This is an H1</h1>
                <p>Followed by a Paragraph with no blank line.</p>
                """
            )
        )

    def test_hash_h2_followed_by_p(self):
        self.assertMarkdownRenders(
            self.dedent(
                """
                ## This is an H2
                Followed by a Paragraph with no blank line.
                """
            ),
            self.dedent(
                """
                <h2>This is an H2</h2>
                <p>Followed by a Paragraph with no blank line.</p>
                """
            )
        )

    def test_hash_h3_followed_by_p(self):
        self.assertMarkdownRenders(
            self.dedent(
                """
                ### This is an H3
                Followed by a Paragraph with no blank line.
                """
            ),
            self.dedent(
                """
                <h3>This is an H3</h3>
                <p>Followed by a Paragraph with no blank line.</p>
                """
            )
        )

    def test_hash_h4_followed_by_p(self):
        self.assertMarkdownRenders(
            self.dedent(
                """
                #### This is an H4
                Followed by a Paragraph with no blank line.
                """
            ),
            self.dedent(
                """
                <h4>This is an H4</h4>
                <p>Followed by a Paragraph with no blank line.</p>
                """
            )
        )

    def test_hash_h5_followed_by_p(self):
        self.assertMarkdownRenders(
            self.dedent(
                """
                ##### This is an H5
                Followed by a Paragraph with no blank line.
                """
            ),
            self.dedent(
                """
                <h5>This is an H5</h5>
                <p>Followed by a Paragraph with no blank line.</p>
                """
            )
        )

    def test_hash_h6_followed_by_p(self):
        self.assertMarkdownRenders(
            self.dedent(
                """
                ###### This is an H6
                Followed by a Paragraph with no blank line.
                """
            ),
            self.dedent(
                """
                <h6>This is an H6</h6>
                <p>Followed by a Paragraph with no blank line.</p>
                """
            )
        )

    def test_hash_h1_leading_space(self):
        self.assertMarkdownRenders(
            ' # This is an H1',

            '<p># This is an H1</p>'
        )

    def test_hash_h2_leading_space(self):
        self.assertMarkdownRenders(
            ' ## This is an H2',

            '<p>## This is an H2</p>'
        )

    def test_hash_h3_leading_space(self):
        self.assertMarkdownRenders(
            ' ### This is an H3',

            '<p>### This is an H3</p>'
        )

    def test_hash_h4_leading_space(self):
        self.assertMarkdownRenders(
            ' #### This is an H4',

            '<p>#### This is an H4</p>'
        )

    def test_hash_h5_leading_space(self):
        self.assertMarkdownRenders(
            ' ##### This is an H5',

            '<p>##### This is an H5</p>'
        )

    def test_hash_h6_leading_space(self):
        self.assertMarkdownRenders(
            ' ###### This is an H6',

            '<p>###### This is an H6</p>'
        )

    def test_hash_h1_open_trailing_space(self):
        self.assertMarkdownRenders(
            '# This is an H1 ',

            '<h1>This is an H1</h1>'
        )

    def test_hash_h2_open_trailing_space(self):
        self.assertMarkdownRenders(
            '## This is an H2 ',

            '<h2>This is an H2</h2>'
        )

    def test_hash_h3_open_trailing_space(self):
        self.assertMarkdownRenders(
            '### This is an H3 ',

            '<h3>This is an H3</h3>'
        )

    def test_hash_h4_open_trailing_space(self):
        self.assertMarkdownRenders(
            '#### This is an H4 ',

            '<h4>This is an H4</h4>'
        )

    def test_hash_h5_open_trailing_space(self):
        self.assertMarkdownRenders(
            '##### This is an H5 ',

            '<h5>This is an H5</h5>'
        )

    def test_hash_h6_open_trailing_space(self):
        self.assertMarkdownRenders(
            '###### This is an H6 ',

            '<h6>This is an H6</h6>'
        )

    def test_hash_gt6_open_trailing_space(self):
        self.assertMarkdownRenders(
            '####### This is an H6 ',

            '<h6># This is an H6</h6>'
        )

    # TODO: Possibly change the following behavior. While this follows the behavior
    # of markdown.pl, it is rather uncommon and not necessarily intuitive.
    # See: https://johnmacfarlane.net/babelmark2/?normalize=1&text=%23+This+is+an+H1+%23+
    def test_hash_h1_closed_trailing_space(self):
        self.assertMarkdownRenders(
            '# This is an H1 # ',

            '<h1>This is an H1 #</h1>'
        )

    def test_hash_h2_closed_trailing_space(self):
        self.assertMarkdownRenders(
            '## This is an H2 ## ',

            '<h2>This is an H2 ##</h2>'
        )

    def test_hash_h3_closed_trailing_space(self):
        self.assertMarkdownRenders(
            '### This is an H3 ### ',

            '<h3>This is an H3 ###</h3>'
        )

    def test_hash_h4_closed_trailing_space(self):
        self.assertMarkdownRenders(
            '#### This is an H4 #### ',

            '<h4>This is an H4 ####</h4>'
        )

    def test_hash_h5_closed_trailing_space(self):
        self.assertMarkdownRenders(
            '##### This is an H5 ##### ',

            '<h5>This is an H5 #####</h5>'
        )

    def test_hash_h6_closed_trailing_space(self):
        self.assertMarkdownRenders(
            '###### This is an H6 ###### ',

            '<h6>This is an H6 ######</h6>'
        )

    def test_hash_gt6_closed_trailing_space(self):
        self.assertMarkdownRenders(
            '####### This is an H6 ####### ',

            '<h6># This is an H6 #######</h6>'
        )

    def test_no_blank_lines_between_hashs(self):
        self.assertMarkdownRenders(
            self.dedent(
                """
                # This is an H1
                ## This is an H2
                """
            ),
            self.dedent(
                """
                <h1>This is an H1</h1>
                <h2>This is an H2</h2>
                """
            )
        )

    def test_random_hash_levels(self):
        self.assertMarkdownRenders(
            self.dedent(
                """
                ### H3
                ###### H6
                # H1
                ##### H5
                #### H4
                ## H2
                ### H3
                """
            ),
            self.dedent(
                """
                <h3>H3</h3>
                <h6>H6</h6>
                <h1>H1</h1>
                <h5>H5</h5>
                <h4>H4</h4>
                <h2>H2</h2>
                <h3>H3</h3>
                """
            )
        )

    def test_hash_followed_by_p(self):
        self.assertMarkdownRenders(
            self.dedent(
                """
                # This is an H1
                Followed by a Paragraph with no blank line.
                """
            ),
            self.dedent(
                """
                <h1>This is an H1</h1>
                <p>Followed by a Paragraph with no blank line.</p>
                """
            )
        )

    def test_p_followed_by_hash(self):
        self.assertMarkdownRenders(
            self.dedent(
                """
                This is a Paragraph.
                # Followed by an H1 with no blank line.
                """
            ),
            self.dedent(
                """
                <p>This is a Paragraph.</p>
                <h1>Followed by an H1 with no blank line.</h1>
                """
            )
        )

    def test_escaped_hash(self):
        self.assertMarkdownRenders(
            "### H3 \\###",
            self.dedent(
                """
                <h3>H3 #</h3>
                """
            )
        )

    def test_unescaped_hash(self):
        self.assertMarkdownRenders(
            "### H3 \\\\###",
            self.dedent(
                """
                <h3>H3 \\</h3>
                """
            )
        )



"""
Python Markdown

A Python implementation of John Gruber's Markdown.

Documentation: https://python-markdown.github.io/
GitHub: https://github.com/Python-Markdown/markdown/
PyPI: https://pypi.org/project/Markdown/

Started by Manfred Stienstra (http://www.dwerg.net/).
Maintained for a few years by Yuri Takhteyev (http://www.freewisdom.org).
Currently maintained by Waylan Limberg (https://github.com/waylan),
Dmitry Shachnev (https://github.com/mitya57) and Isaac Muse (https://github.com/facelessuser).

Copyright 2007-2023 The Python Markdown Project (v. 1.7 and later)
Copyright 2004, 2005, 2006 Yuri Takhteyev (v. 0.2-1.6b)
Copyright 2004 Manfred Stienstra (the original version)

License: BSD (see LICENSE.md for details).

Python-Markdown Extension Regression Tests
==========================================

A collection of regression tests to confirm that the included extensions
continue to work as advertised. This used to be accomplished by `doctests`.
"""

import unittest
import markdown


class TestExtensionClass(unittest.TestCase):
    """ Test markdown.extensions.Extension. """

    def setUp(self):
        class TestExtension(markdown.extensions.Extension):
            config = {
                'foo': ['bar', 'Description of foo'],
                'bar': ['baz', 'Description of bar']
            }

        self.ext = TestExtension()
        self.ExtKlass = TestExtension

    def testGetConfig(self):
        self.assertEqual(self.ext.getConfig('foo'), 'bar')

    def testGetConfigDefault(self):
        self.assertEqual(self.ext.getConfig('baz'), '')
        self.assertEqual(self.ext.getConfig('baz', default='missing'), 'missing')

    def testGetConfigs(self):
        self.assertEqual(self.ext.getConfigs(), {'foo': 'bar', 'bar': 'baz'})

    def testGetConfigInfo(self):
        self.assertEqual(
            dict(self.ext.getConfigInfo()),
            dict([
                ('foo', 'Description of foo'),
                ('bar', 'Description of bar')
            ])
        )

    def testSetConfig(self):
        self.ext.setConfig('foo', 'baz')
        self.assertEqual(self.ext.getConfigs(), {'foo': 'baz', 'bar': 'baz'})

    def testSetConfigWithBadKey(self):
        # `self.ext.setConfig('bad', 'baz)` => `KeyError`
        self.assertRaises(KeyError, self.ext.setConfig, 'bad', 'baz')

    def testConfigAsKwargsOnInit(self):
        ext = self.ExtKlass(foo='baz', bar='blah')
        self.assertEqual(ext.getConfigs(), {'foo': 'baz', 'bar': 'blah'})


class TestMetaData(unittest.TestCase):
    """ Test `MetaData` extension. """

    def setUp(self):
        self.md = markdown.Markdown(extensions=['meta'])

    def testBasicMetaData(self):
        """ Test basic metadata. """

        text = '''Title: A Test Doc.
Author: Waylan Limberg
        John Doe
Blank_Data:

The body. This is paragraph one.'''
        self.assertEqual(
            self.md.convert(text),
            '<p>The body. This is paragraph one.</p>'
        )
        self.assertEqual(
            self.md.Meta, {
                'author': ['Waylan Limberg', 'John Doe'],
                'blank_data': [''],
                'title': ['A Test Doc.']
            }
        )

    def testYamlMetaData(self):
        """ Test metadata specified as simple YAML. """

        text = '''---
Title: A Test Doc.
Author: [Waylan Limberg, John Doe]
Blank_Data:
---

The body. This is paragraph one.'''
        self.assertEqual(
            self.md.convert(text),
            '<p>The body. This is paragraph one.</p>'
        )
        self.assertEqual(
            self.md.Meta, {
                'author': ['[Waylan Limberg, John Doe]'],
                'blank_data': [''],
                'title': ['A Test Doc.']
            }
        )

    def testMissingMetaData(self):
        """ Test document without Meta Data. """

        text = '    Some Code - not extra lines of meta data.'
        self.assertEqual(
            self.md.convert(text),
            '<pre><code>Some Code - not extra lines of meta data.\n'
            '</code></pre>'
        )
        self.assertEqual(self.md.Meta, {})

    def testMetaDataWithoutNewline(self):
        """ Test document with only metadata and no newline at end."""
        text = 'title: No newline'
        self.assertEqual(self.md.convert(text), '')
        self.assertEqual(self.md.Meta, {'title': ['No newline']})

    def testMetaDataReset(self):
        """ Test that reset call remove Meta entirely """

        text = '''Title: A Test Doc.
Author: Waylan Limberg
        John Doe
Blank_Data:

The body. This is paragraph one.'''
        self.md.convert(text)

        self.md.reset()
        self.assertEqual(self.md.Meta, {})


class TestWikiLinks(unittest.TestCase):
    """ Test `Wikilinks` Extension. """

    def setUp(self):
        self.md = markdown.Markdown(extensions=['wikilinks'])
        self.text = "Some text with a [[WikiLink]]."

    def testBasicWikilinks(self):
        """ Test `[[wikilinks]]`. """

        self.assertEqual(
            self.md.convert(self.text),
            '<p>Some text with a '
            '<a class="wikilink" href="/WikiLink/">WikiLink</a>.</p>'
        )

    def testWikilinkWhitespace(self):
        """ Test whitespace in `wikilinks`. """
        self.assertEqual(
            self.md.convert('[[ foo bar_baz ]]'),
            '<p><a class="wikilink" href="/foo_bar_baz/">foo bar_baz</a></p>'
        )
        self.assertEqual(
            self.md.convert('foo [[ ]] bar'),
            '<p>foo  bar</p>'
        )

    def testSimpleSettings(self):
        """ Test Simple Settings. """

        self.assertEqual(markdown.markdown(
            self.text, extensions=[
                markdown.extensions.wikilinks.WikiLinkExtension(
                    base_url='/wiki/',
                    end_url='.html',
                    html_class='foo')
                ]
            ),
            '<p>Some text with a '
            '<a class="foo" href="/wiki/WikiLink.html">WikiLink</a>.</p>')

    def testComplexSettings(self):
        """ Test Complex Settings. """

        md = markdown.Markdown(
            extensions=['wikilinks'],
            extension_configs={
                'wikilinks': [
                    ('base_url', 'http://example.com/'),
                    ('end_url', '.html'),
                    ('html_class', '')
                ]
            },
            safe_mode=True
        )
        self.assertEqual(
            md.convert(self.text),
            '<p>Some text with a '
            '<a href="http://example.com/WikiLink.html">WikiLink</a>.</p>'
        )

    def testWikilinksMetaData(self):
        """ test `MetaData` with `Wikilinks` Extension. """

        text = """wiki_base_url: http://example.com/
wiki_end_url:   .html
wiki_html_class:

Some text with a [[WikiLink]]."""
        md = markdown.Markdown(extensions=['meta', 'wikilinks'])
        self.assertEqual(
            md.convert(text),
            '<p>Some text with a '
            '<a href="http://example.com/WikiLink.html">WikiLink</a>.</p>'
        )

        # `MetaData` should not carry over to next document:
        self.assertEqual(
            md.convert("No [[MetaData]] here."),
            '<p>No <a class="wikilink" href="/MetaData/">MetaData</a> '
            'here.</p>'
        )

    def testURLCallback(self):
        """ Test used of a custom URL builder. """

        from markdown.extensions.wikilinks import WikiLinkExtension

        def my_url_builder(label, base, end):
            return '/bar/'

        md = markdown.Markdown(extensions=[WikiLinkExtension(build_url=my_url_builder)])
        self.assertEqual(
            md.convert('[[foo]]'),
            '<p><a class="wikilink" href="/bar/">foo</a></p>'
        )


class TestAdmonition(unittest.TestCase):
    """ Test Admonition Extension. """

    def setUp(self):
        self.md = markdown.Markdown(extensions=['admonition'])

    def testRE(self):
        RE = self.md.parser.blockprocessors['admonition'].RE
        tests = [
            ('!!! note', ('note', None)),
            ('!!! note "Please Note"', ('note', 'Please Note')),
            ('!!! note ""', ('note', '')),
        ]
        for test, expected in tests:
            self.assertEqual(RE.match(test).groups(), expected)


class TestSmarty(unittest.TestCase):
    def setUp(self):
        config = {
            'smarty': [
                ('smart_angled_quotes', True),
                ('substitutions', {
                    'ndash': '\u2013',
                    'mdash': '\u2014',
                    'ellipsis': '\u2026',
                    'left-single-quote': '&sbquo;',  # `sb` is not a typo!
                    'right-single-quote': '&lsquo;',
                    'left-double-quote': '&bdquo;',
                    'right-double-quote': '&ldquo;',
                    'left-angle-quote': '[',
                    'right-angle-quote': ']',
                }),
            ]
        }
        self.md = markdown.Markdown(
            extensions=['smarty'],
            extension_configs=config
        )

    def testCustomSubstitutions(self):
        text = """<< The "Unicode char of the year 2014"
is the 'mdash': ---
Must not be confused with 'ndash'  (--) ... >>
"""
        correct = """<p>[ The &bdquo;Unicode char of the year 2014&ldquo;
is the &sbquo;mdash&lsquo;: \u2014
Must not be confused with &sbquo;ndash&lsquo;  (\u2013) \u2026 ]</p>"""
        self.assertEqual(self.md.convert(text), correct)



"""
Python Markdown

A Python implementation of John Gruber's Markdown.

Documentation: https://python-markdown.github.io/
GitHub: https://github.com/Python-Markdown/markdown/
PyPI: https://pypi.org/project/Markdown/

Started by Manfred Stienstra (http://www.dwerg.net/).
Maintained for a few years by Yuri Takhteyev (http://www.freewisdom.org).
Currently maintained by Waylan Limberg (https://github.com/waylan),
Dmitry Shachnev (https://github.com/mitya57) and Isaac Muse (https://github.com/facelessuser).

Copyright 2007-2019 The Python Markdown Project (v. 1.7 and later)
Copyright 2004, 2005, 2006 Yuri Takhteyev (v. 0.2-1.6b)
Copyright 2004 Manfred Stienstra (the original version)

License: BSD (see LICENSE.md for details).
"""

from markdown.test_tools import TestCase
from markdown.extensions.codehilite import CodeHiliteExtension, CodeHilite
from markdown import extensions, treeprocessors
import os
import xml.etree.ElementTree as etree

try:
    import pygments  # noqa
    has_pygments = True
except ImportError:
    has_pygments = False

# The version required by the tests is the version specified and installed in the `pygments` tox environment.
# In any environment where the `PYGMENTS_VERSION` environment variable is either not defined or doesn't
# match the version of Pygments installed, all tests which rely in Pygments will be skipped.
required_pygments_version = os.environ.get('PYGMENTS_VERSION', '')


class TestCodeHiliteClass(TestCase):
    """ Test the markdown.extensions.codehilite.CodeHilite class. """

    def setUp(self):
        if has_pygments and pygments.__version__ != required_pygments_version:
            self.skipTest(f'Pygments=={required_pygments_version} is required')

    maxDiff = None

    def assertOutputEquals(self, source, expected, **options):
        """
        Test that source code block results in the expected output with given options.
        """

        output = CodeHilite(source, **options).hilite()
        self.assertMultiLineEqual(output.strip(), expected)

    def test_codehilite_defaults(self):
        if has_pygments:
            # Odd result as no `lang` given and a single comment is not enough for guessing.
            expected = (
                '<div class="codehilite"><pre><span></span><code><span class="err"># A Code Comment</span>\n'
                '</code></pre></div>'
            )
        else:
            expected = (
                '<pre class="codehilite"><code># A Code Comment\n'
                '</code></pre>'
            )
        self.assertOutputEquals('# A Code Comment', expected)

    def test_codehilite_guess_lang(self):
        if has_pygments:
            expected = (
                '<div class="codehilite"><pre><span></span><code><span class="cp">&lt;?php</span> '
                '<span class="k">print</span><span class="p">(</span><span class="s2">&quot;Hello World&quot;</span>'
                '<span class="p">);</span> <span class="cp">?&gt;</span>\n'
                '</code></pre></div>'
            )
        else:
            expected = (
                '<pre class="codehilite"><code>&lt;?php print(&quot;Hello World&quot;); ?&gt;\n'
                '</code></pre>'
            )
        # Use PHP as the the starting `<?php` tag ensures an accurate guess.
        self.assertOutputEquals('<?php print("Hello World"); ?>', expected, guess_lang=True)

    def test_codehilite_guess_lang_plain_text(self):
        if has_pygments:
            expected = (
                '<div class="codehilite"><pre><span></span><code><span class="err">plain text</span>\n'
                '</code></pre></div>'
            )
        else:
            expected = (
                '<pre class="codehilite"><code>plain text\n'
                '</code></pre>'
            )
        # This will be difficult to guess.
        self.assertOutputEquals('plain text', expected, guess_lang=True)

    def test_codehilite_set_lang(self):
        if has_pygments:
            # Note an extra `<span class="x">` is added to end of code block when `lang` explicitly set.
            # Compare with expected output for `test_guess_lang`. Not sure why this happens.
            expected = (
                '<div class="codehilite"><pre><span></span><code><span class="cp">&lt;?php</span> '
                '<span class="k">print</span><span class="p">(</span><span class="s2">&quot;Hello World&quot;</span>'
                '<span class="p">);</span> <span class="cp">?&gt;</span><span class="x"></span>\n'
                '</code></pre></div>'
            )
        else:
            expected = (
                '<pre class="codehilite"><code class="language-php">&lt;?php print(&quot;Hello World&quot;); ?&gt;\n'
                '</code></pre>'
            )
        self.assertOutputEquals('<?php print("Hello World"); ?>', expected, lang='php')

    def test_codehilite_bad_lang(self):
        if has_pygments:
            expected = (
                '<div class="codehilite"><pre><span></span><code><span class="cp">&lt;?php</span> '
                '<span class="k">print</span><span class="p">(</span><span class="s2">'
                '&quot;Hello World&quot;</span><span class="p">);</span> <span class="cp">?&gt;</span>\n'
                '</code></pre></div>'
            )
        else:
            # Note that without Pygments there is no way to check that the language name is bad.
            expected = (
                '<pre class="codehilite"><code class="language-unkown">'
                '&lt;?php print(&quot;Hello World&quot;); ?&gt;\n'
                '</code></pre>'
            )
        # The starting `<?php` tag ensures an accurate guess.
        self.assertOutputEquals('<?php print("Hello World"); ?>', expected, lang='unkown')

    def test_codehilite_use_pygments_false(self):
        expected = (
            '<pre class="codehilite"><code class="language-php">&lt;?php print(&quot;Hello World&quot;); ?&gt;\n'
            '</code></pre>'
        )
        self.assertOutputEquals('<?php print("Hello World"); ?>', expected, lang='php', use_pygments=False)

    def test_codehilite_lang_prefix_empty(self):
        expected = (
            '<pre class="codehilite"><code class="php">&lt;?php print(&quot;Hello World&quot;); ?&gt;\n'
            '</code></pre>'
        )
        self.assertOutputEquals(
            '<?php print("Hello World"); ?>', expected, lang='php', use_pygments=False, lang_prefix=''
        )

    def test_codehilite_lang_prefix(self):
        expected = (
            '<pre class="codehilite"><code class="lang-php">&lt;?php print(&quot;Hello World&quot;); ?&gt;\n'
            '</code></pre>'
        )
        self.assertOutputEquals(
            '<?php print("Hello World"); ?>', expected, lang='php', use_pygments=False, lang_prefix='lang-'
        )

    def test_codehilite_linenos_true(self):
        if has_pygments:
            expected = (
                '<table class="codehilitetable"><tr><td class="linenos"><div class="linenodiv"><pre>1</pre></div>'
                '</td><td class="code"><div class="codehilite"><pre><span></span><code>plain text\n'
                '</code></pre></div>\n'
                '</td></tr></table>'
            )
        else:
            expected = (
                '<pre class="codehilite"><code class="language-text linenums">plain text\n'
                '</code></pre>'
            )
        self.assertOutputEquals('plain text', expected, lang='text', linenos=True)

    def test_codehilite_linenos_false(self):
        if has_pygments:
            expected = (
                '<div class="codehilite"><pre><span></span><code>plain text\n'
                '</code></pre></div>'
            )
        else:
            expected = (
                '<pre class="codehilite"><code class="language-text">plain text\n'
                '</code></pre>'
            )
        self.assertOutputEquals('plain text', expected, lang='text', linenos=False)

    def test_codehilite_linenos_none(self):
        if has_pygments:
            expected = (
                '<div class="codehilite"><pre><span></span><code>plain text\n'
                '</code></pre></div>'
            )
        else:
            expected = (
                '<pre class="codehilite"><code class="language-text">plain text\n'
                '</code></pre>'
            )
        self.assertOutputEquals('plain text', expected, lang='text', linenos=None)

    def test_codehilite_linenos_table(self):
        if has_pygments:
            expected = (
                '<table class="codehilitetable"><tr><td class="linenos"><div class="linenodiv"><pre>1</pre></div>'
                '</td><td class="code"><div class="codehilite"><pre><span></span><code>plain text\n'
                '</code></pre></div>\n'
                '</td></tr></table>'
            )
        else:
            expected = (
                '<pre class="codehilite"><code class="language-text linenums">plain text\n'
                '</code></pre>'
            )
        self.assertOutputEquals('plain text', expected, lang='text', linenos='table')

    def test_codehilite_linenos_inline(self):
        if has_pygments:
            expected = (
                '<div class="codehilite"><pre><span></span><code><span class="linenos">1</span>plain text\n'
                '</code></pre></div>'
            )
        else:
            expected = (
                '<pre class="codehilite"><code class="language-text linenums">plain text\n'
                '</code></pre>'
            )
        self.assertOutputEquals('plain text', expected, lang='text', linenos='inline')

    def test_codehilite_linenums_true(self):
        if has_pygments:
            expected = (
                '<table class="codehilitetable"><tr><td class="linenos"><div class="linenodiv"><pre>1</pre></div>'
                '</td><td class="code"><div class="codehilite"><pre><span></span><code>plain text\n'
                '</code></pre></div>\n'
                '</td></tr></table>'
            )
        else:
            expected = (
                '<pre class="codehilite"><code class="language-text linenums">plain text\n'
                '</code></pre>'
            )
        self.assertOutputEquals('plain text', expected, lang='text', linenums=True)

    def test_codehilite_set_cssclass(self):
        if has_pygments:
            expected = (
                '<div class="override"><pre><span></span><code>plain text\n'
                '</code></pre></div>'
            )
        else:
            expected = (
                '<pre class="override"><code class="language-text">plain text\n'
                '</code></pre>'
            )
        self.assertOutputEquals('plain text', expected, lang='text', cssclass='override')

    def test_codehilite_set_css_class(self):
        if has_pygments:
            expected = (
                '<div class="override"><pre><span></span><code>plain text\n'
                '</code></pre></div>'
            )
        else:
            expected = (
                '<pre class="override"><code class="language-text">plain text\n'
                '</code></pre>'
            )
        self.assertOutputEquals('plain text', expected, lang='text', css_class='override')

    def test_codehilite_linenostart(self):
        if has_pygments:
            expected = (
                '<div class="codehilite"><pre><span></span><code><span class="linenos">42</span>plain text\n'
                '</code></pre></div>'
            )
        else:
            # TODO: Implement `linenostart` for no-Pygments. Will need to check what JavaScript libraries look for.
            expected = (
                '<pre class="codehilite"><code class="language-text linenums">plain text\n'
                '</code></pre>'
            )
        self.assertOutputEquals('plain text', expected, lang='text', linenos='inline', linenostart=42)

    def test_codehilite_linenos_hl_lines(self):
        if has_pygments:
            expected = (
                '<div class="codehilite"><pre><span></span><code>'
                '<span class="linenos">1</span><span class="hll">line 1\n'
                '</span><span class="linenos">2</span>line 2\n'
                '<span class="linenos">3</span><span class="hll">line 3\n'
                '</span></code></pre></div>'
            )
        else:
            expected = (
                '<pre class="codehilite"><code class="language-text linenums">line 1\n'
                'line 2\n'
                'line 3\n'
                '</code></pre>'
            )
        self.assertOutputEquals('line 1\nline 2\nline 3', expected, lang='text', linenos='inline', hl_lines=[1, 3])

    def test_codehilite_linenos_linenostep(self):
        if has_pygments:
            expected = (
                '<div class="codehilite"><pre><span></span><code><span class="linenos"> </span>line 1\n'
                '<span class="linenos">2</span>line 2\n'
                '<span class="linenos"> </span>line 3\n'
                '</code></pre></div>'
            )
        else:
            expected = (
                '<pre class="codehilite"><code class="language-text linenums">line 1\n'
                'line 2\n'
                'line 3\n'
                '</code></pre>'
            )
        self.assertOutputEquals('line 1\nline 2\nline 3', expected, lang='text', linenos='inline', linenostep=2)

    def test_codehilite_linenos_linenospecial(self):
        if has_pygments:
            expected = (
                '<div class="codehilite"><pre><span></span><code><span class="linenos">1</span>line 1\n'
                '<span class="linenos special">2</span>line 2\n'
                '<span class="linenos">3</span>line 3\n'
                '</code></pre></div>'
            )
        else:
            expected = (
                '<pre class="codehilite"><code class="language-text linenums">line 1\n'
                'line 2\n'
                'line 3\n'
                '</code></pre>'
            )
        self.assertOutputEquals('line 1\nline 2\nline 3', expected, lang='text', linenos='inline', linenospecial=2)

    def test_codehilite_startinline(self):
        if has_pygments:
            expected = (
                '<div class="codehilite"><pre><span></span><code><span class="k">print</span><span class="p">(</span>'
                '<span class="s2">&quot;Hello World&quot;</span><span class="p">);</span>\n'
                '</code></pre></div>'
            )
        else:
            expected = (
                '<pre class="codehilite"><code class="language-php">print(&quot;Hello World&quot;);\n'
                '</code></pre>'
            )
        self.assertOutputEquals('print("Hello World");', expected, lang='php', startinline=True)


class TestCodeHiliteExtension(TestCase):
    """ Test codehilite extension. """

    def setUp(self):
        if has_pygments and pygments.__version__ != required_pygments_version:
            self.skipTest(f'Pygments=={required_pygments_version} is required')

        # Define a custom Pygments formatter (same example in the documentation)
        if has_pygments:
            class CustomAddLangHtmlFormatter(pygments.formatters.HtmlFormatter):
                def __init__(self, lang_str='', **options):
                    super().__init__(**options)
                    self.lang_str = lang_str

                def _wrap_code(self, source):
                    yield 0, f'<code class="{self.lang_str}">'
                    yield from source
                    yield 0, '</code>'
        else:
            CustomAddLangHtmlFormatter = None

        self.custom_pygments_formatter = CustomAddLangHtmlFormatter

    maxDiff = None

    def testBasicCodeHilite(self):
        if has_pygments:
            # Odd result as no `lang` given and a single comment is not enough for guessing.
            expected = (
                '<div class="codehilite"><pre><span></span><code><span class="err"># A Code Comment</span>\n'
                '</code></pre></div>'
            )
        else:
            expected = (
                '<pre class="codehilite"><code># A Code Comment\n'
                '</code></pre>'
            )
        self.assertMarkdownRenders(
            '\t# A Code Comment',
            expected,
            extensions=['codehilite']
        )

    def testLinenumsTrue(self):
        if has_pygments:
            expected = (
                '<table class="codehilitetable"><tr>'
                '<td class="linenos"><div class="linenodiv"><pre>1</pre></div></td>'
                '<td class="code"><div class="codehilite"><pre><span></span>'
                '<code><span class="err"># A Code Comment</span>\n'
                '</code></pre></div>\n'
                '</td></tr></table>'
            )
        else:
            expected = (
                '<pre class="codehilite"><code class="linenums"># A Code Comment\n'
                '</code></pre>'
            )
        self.assertMarkdownRenders(
            '\t# A Code Comment',
            expected,
            extensions=[CodeHiliteExtension(linenums=True)]
        )

    def testLinenumsFalse(self):
        if has_pygments:
            expected = (
                '<div class="codehilite"><pre><span></span><code><span class="c1"># A Code Comment</span>\n'
                '</code></pre></div>'
            )
        else:
            expected = (
                '<pre class="codehilite"><code class="language-python"># A Code Comment\n'
                '</code></pre>'
            )
        self.assertMarkdownRenders(
            (
                '\t#!Python\n'
                '\t# A Code Comment'
            ),
            expected,
            extensions=[CodeHiliteExtension(linenums=False)]
        )

    def testLinenumsNone(self):
        if has_pygments:
            expected = (
                '<div class="codehilite"><pre><span></span><code><span class="err"># A Code Comment</span>\n'
                '</code></pre></div>'
            )
        else:
            expected = (
                '<pre class="codehilite"><code># A Code Comment\n'
                '</code></pre>'
            )
        self.assertMarkdownRenders(
            '\t# A Code Comment',
            expected,
            extensions=[CodeHiliteExtension(linenums=None)]
        )

    def testLinenumsNoneWithShebang(self):
        if has_pygments:
            expected = (
                '<table class="codehilitetable"><tr>'
                '<td class="linenos"><div class="linenodiv"><pre>1</pre></div></td>'
                '<td class="code"><div class="codehilite"><pre><span></span>'
                '<code><span class="c1"># A Code Comment</span>\n'
                '</code></pre></div>\n'
                '</td></tr></table>'
            )
        else:
            expected = (
                '<pre class="codehilite"><code class="language-python linenums"># A Code Comment\n'
                '</code></pre>'
            )
        self.assertMarkdownRenders(
            (
                '\t#!Python\n'
                '\t# A Code Comment'
            ),
            expected,
            extensions=[CodeHiliteExtension(linenums=None)]
        )

    def testLinenumsNoneWithColon(self):
        if has_pygments:
            expected = (
                '<div class="codehilite"><pre><span></span><code><span class="c1"># A Code Comment</span>\n'
                '</code></pre></div>'
            )
        else:
            expected = (
                '<pre class="codehilite"><code class="language-python"># A Code Comment\n'
                '</code></pre>'
            )
        self.assertMarkdownRenders(
            (
                '\t:::Python\n'
                '\t# A Code Comment'
            ),
            expected,
            extensions=[CodeHiliteExtension(linenums=None)]
        )

    def testHighlightLinesWithColon(self):
        if has_pygments:
            expected = (
                '<div class="codehilite"><pre><span></span><code><span class="hll"><span class="c1">#line 1</span>\n'
                '</span><span class="c1">#line 2</span>\n'
                '<span class="c1">#line 3</span>\n'
                '</code></pre></div>'
            )
        else:
            expected = (
                '<pre class="codehilite"><code class="language-python">#line 1\n'
                '#line 2\n'
                '#line 3\n'
                '</code></pre>'
            )
        # Double quotes
        self.assertMarkdownRenders(
            (
                '\t:::Python hl_lines="1"\n'
                '\t#line 1\n'
                '\t#line 2\n'
                '\t#line 3'
            ),
            expected,
            extensions=['codehilite']
        )
        # Single quotes
        self.assertMarkdownRenders(
            (
                "\t:::Python hl_lines='1'\n"
                '\t#line 1\n'
                '\t#line 2\n'
                '\t#line 3'
            ),
            expected,
            extensions=['codehilite']
        )

    def testUsePygmentsFalse(self):
        self.assertMarkdownRenders(
            (
                '\t:::Python\n'
                '\t# A Code Comment'
            ),
            (
                '<pre class="codehilite"><code class="language-python"># A Code Comment\n'
                '</code></pre>'
            ),
            extensions=[CodeHiliteExtension(use_pygments=False)]
        )

    def testLangPrefixEmpty(self):
        self.assertMarkdownRenders(
            (
                '\t:::Python\n'
                '\t# A Code Comment'
            ),
            (
                '<pre class="codehilite"><code class="python"># A Code Comment\n'
                '</code></pre>'
            ),
            extensions=[CodeHiliteExtension(use_pygments=False, lang_prefix='')]
        )

    def testLangPrefix(self):
        self.assertMarkdownRenders(
            (
                '\t:::Python\n'
                '\t# A Code Comment'
            ),
            (
                '<pre class="codehilite"><code class="lang-python"># A Code Comment\n'
                '</code></pre>'
            ),
            extensions=[CodeHiliteExtension(use_pygments=False, lang_prefix='lang-')]
        )

    def testDoubleEscape(self):
        if has_pygments:
            expected = (
                '<div class="codehilite"><pre>'
                '<span></span>'
                '<code><span class="p">&lt;</span><span class="nt">span</span><span class="p">&gt;</span>'
                'This<span class="ni">&amp;amp;</span>That'
                '<span class="p">&lt;/</span><span class="nt">span</span><span class="p">&gt;</span>'
                '\n</code></pre></div>'
            )
        else:
            expected = (
                '<pre class="codehilite"><code class="language-html">'
                '&lt;span&gt;This&amp;amp;That&lt;/span&gt;\n'
                '</code></pre>'
            )
        self.assertMarkdownRenders(
            (
                '\t:::html\n'
                '\t<span>This&amp;That</span>'
            ),
            expected,
            extensions=['codehilite']
        )

    def testEntitiesIntact(self):
        if has_pygments:
            expected = (
                '<div class="codehilite"><pre>'
                '<span></span>'
                '<code>&lt; &amp;lt; and &gt; &amp;gt;'
                '\n</code></pre></div>'
            )
        else:
            expected = (
                '<pre class="codehilite"><code class="language-text">'
                '&lt; &amp;lt; and &gt; &amp;gt;\n'
                '</code></pre>'
            )
        self.assertMarkdownRenders(
            (
                '\t:::text\n'
                '\t< &lt; and > &gt;'
            ),
            expected,
            extensions=['codehilite']
        )

    def testHighlightAmps(self):
        if has_pygments:
            expected = (
                '<div class="codehilite"><pre><span></span><code>&amp;\n'
                '&amp;amp;\n'
                '&amp;amp;amp;\n'
                '</code></pre></div>'
            )
        else:
            expected = (
                '<pre class="codehilite"><code class="language-text">&amp;\n'
                '&amp;amp;\n'
                '&amp;amp;amp;\n'
                '</code></pre>'
            )
        self.assertMarkdownRenders(
            (
                '\t:::text\n'
                '\t&\n'
                '\t&amp;\n'
                '\t&amp;amp;'
            ),
            expected,
            extensions=['codehilite']
        )

    def testUnknownOption(self):
        if has_pygments:
            # Odd result as no `lang` given and a single comment is not enough for guessing.
            expected = (
                '<div class="codehilite"><pre><span></span><code><span class="err"># A Code Comment</span>\n'
                '</code></pre></div>'
            )
        else:
            expected = (
                '<pre class="codehilite"><code># A Code Comment\n'
                '</code></pre>'
            )
        self.assertMarkdownRenders(
            '\t# A Code Comment',
            expected,
            extensions=[CodeHiliteExtension(unknown='some value')],
        )

    def testMultipleBlocksSameStyle(self):
        if has_pygments:
            # See also: https://github.com/Python-Markdown/markdown/issues/1240
            expected = (
                '<div class="codehilite" style="background: #202020"><pre style="line-height: 125%; margin: 0;">'
                '<span></span><code><span style="color: #999999; font-style: italic"># First Code Block</span>\n'
                '</code></pre></div>\n\n'
                '<p>Normal paragraph</p>\n'
                '<div class="codehilite" style="background: #202020"><pre style="line-height: 125%; margin: 0;">'
                '<span></span><code><span style="color: #999999; font-style: italic"># Second Code Block</span>\n'
                '</code></pre></div>'
            )
        else:
            expected = (
                '<pre class="codehilite"><code class="language-python"># First Code Block\n'
                '</code></pre>\n\n'
                '<p>Normal paragraph</p>\n'
                '<pre class="codehilite"><code class="language-python"># Second Code Block\n'
                '</code></pre>'
            )
        self.assertMarkdownRenders(
            (
                '\t:::Python\n'
                '\t# First Code Block\n\n'
                'Normal paragraph\n\n'
                '\t:::Python\n'
                '\t# Second Code Block'
            ),
            expected,
            extensions=[CodeHiliteExtension(pygments_style="native", noclasses=True)]
        )

    def testFormatterLangStr(self):
        if has_pygments:
            expected = (
                '<div class="codehilite"><pre><span></span><code class="language-python">'
                '<span class="c1"># A Code Comment</span>\n'
                '</code></pre></div>'
            )
        else:
            expected = (
                '<pre class="codehilite"><code class="language-python"># A Code Comment\n'
                '</code></pre>'
            )

        self.assertMarkdownRenders(
            '\t:::Python\n'
            '\t# A Code Comment',
            expected,
            extensions=[
                CodeHiliteExtension(
                    guess_lang=False,
                    pygments_formatter=self.custom_pygments_formatter
                )
            ]
        )

    def testFormatterLangStrGuessLang(self):
        if has_pygments:
            expected = (
                '<div class="codehilite"><pre><span></span>'
                '<code class="language-js+php"><span class="cp">&lt;?php</span> '
                '<span class="k">print</span><span class="p">(</span>'
                '<span class="s2">&quot;Hello World&quot;</span>'
                '<span class="p">);</span> <span class="cp">?&gt;</span>\n'
                '</code></pre></div>'
            )
        else:
            expected = (
                '<pre class="codehilite"><code>&lt;?php print(&quot;Hello World&quot;); ?&gt;\n'
                '</code></pre>'
            )
        # Use PHP as the the starting `<?php` tag ensures an accurate guess.
        self.assertMarkdownRenders(
            '\t<?php print("Hello World"); ?>',
            expected,
            extensions=[CodeHiliteExtension(pygments_formatter=self.custom_pygments_formatter)]
        )

    def testFormatterLangStrEmptyLang(self):
        if has_pygments:
            expected = (
                '<div class="codehilite"><pre><span></span>'
                '<code class="language-text"># A Code Comment\n'
                '</code></pre></div>'
            )
        else:
            expected = (
                '<pre class="codehilite"><code># A Code Comment\n'
                '</code></pre>'
            )
        self.assertMarkdownRenders(
            '\t# A Code Comment',
            expected,
            extensions=[
                CodeHiliteExtension(
                    guess_lang=False,
                    pygments_formatter=self.custom_pygments_formatter,
                )
            ]
        )

    def testDoesntCrashWithEmptyCodeTag(self):
        expected = '<h1>Hello</h1>\n<pre><code></code></pre>'
        self.assertMarkdownRenders(
            '# Hello',
            expected,
            extensions=[CodeHiliteExtension(), _ExtensionThatAddsAnEmptyCodeTag()]
        )


class _ExtensionThatAddsAnEmptyCodeTag(extensions.Extension):
    def extendMarkdown(self, md):
        md.treeprocessors.register(_AddCodeTagTreeprocessor(), 'add-code-tag', 40)


class _AddCodeTagTreeprocessor(treeprocessors.Treeprocessor):
    def run(self, root: etree.Element):
        pre = etree.SubElement(root, 'pre')
        etree.SubElement(pre, 'code')



"""
Python Markdown

A Python implementation of John Gruber's Markdown.

Documentation: https://python-markdown.github.io/
GitHub: https://github.com/Python-Markdown/markdown/
PyPI: https://pypi.org/project/Markdown/

Started by Manfred Stienstra (http://www.dwerg.net/).
Maintained for a few years by Yuri Takhteyev (http://www.freewisdom.org).
Currently maintained by Waylan Limberg (https://github.com/waylan),
Dmitry Shachnev (https://github.com/mitya57) and Isaac Muse (https://github.com/facelessuser).

Copyright 2007-2021 The Python Markdown Project (v. 1.7 and later)
Copyright 2004, 2005, 2006 Yuri Takhteyev (v. 0.2-1.6b)
Copyright 2004 Manfred Stienstra (the original version)

License: BSD (see LICENSE.md for details).
"""

from markdown.test_tools import TestCase


class TestAutomaticLinks(TestCase):

    def test_email_address(self):
        self.assertMarkdownRenders(
            'asdfasdfadsfasd <yuri@freewisdom.org> or you can say ',
            '<p>asdfasdfadsfasd <a href="&#109;&#97;&#105;&#108;&#116;&#111;&#58;&#121;&#117;&#114;'
            '&#105;&#64;&#102;&#114;&#101;&#101;&#119;&#105;&#115;&#100;&#111;&#109;&#46;&#111;&#114;'
            '&#103;">&#121;&#117;&#114;&#105;&#64;&#102;&#114;&#101;&#101;&#119;&#105;&#115;&#100;'
            '&#111;&#109;&#46;&#111;&#114;&#103;</a> or you can say </p>'
        )

    def test_mailto_email_address(self):
        self.assertMarkdownRenders(
            'instead <mailto:yuri@freewisdom.org>',
            '<p>instead <a href="&#109;&#97;&#105;&#108;&#116;&#111;&#58;&#121;&#117;&#114;&#105;&#64;'
            '&#102;&#114;&#101;&#101;&#119;&#105;&#115;&#100;&#111;&#109;&#46;&#111;&#114;&#103;">'
            '&#121;&#117;&#114;&#105;&#64;&#102;&#114;&#101;&#101;&#119;&#105;&#115;&#100;&#111;&#109;'
            '&#46;&#111;&#114;&#103;</a></p>'
        )

    def test_email_address_with_ampersand(self):
        self.assertMarkdownRenders(
            '<bob&sue@example.com>',
            '<p><a href="&#109;&#97;&#105;&#108;&#116;&#111;&#58;&#98;&#111;&#98;&#38;&#115;&#117;&#101;'
            '&#64;&#101;&#120;&#97;&#109;&#112;&#108;&#101;&#46;&#99;&#111;&#109;">&#98;&#111;&#98;&amp;'
            '&#115;&#117;&#101;&#64;&#101;&#120;&#97;&#109;&#112;&#108;&#101;&#46;&#99;&#111;&#109;</a></p>'
        )

    def test_invalid_email_address_local_part(self):
        self.assertMarkdownRenders(
            'Missing local-part <@domain>',
            '<p>Missing local-part &lt;@domain&gt;</p>'
        )

    def test_invalid_email_address_domain(self):
        self.assertMarkdownRenders(
            'Missing domain <local-part@>',
            '<p>Missing domain &lt;local-part@&gt;</p>'
        )



"""Generate the code reference pages and navigation."""

import textwrap
import yaml
from pathlib import Path

import mkdocs_gen_files

nav = mkdocs_gen_files.Nav()

per_module_options = {
    "markdown": {"summary": {"attributes": True, "functions": True, "classes": True}}
}

base_path = Path(__file__).resolve().parent.parent

modules = [
    base_path.joinpath("markdown", "__init__.py"),
    base_path.joinpath("markdown", "preprocessors.py"),
    base_path.joinpath("markdown", "blockparser.py"),
    base_path.joinpath("markdown", "blockprocessors.py"),
    base_path.joinpath("markdown", "treeprocessors.py"),
    base_path.joinpath("markdown", "inlinepatterns.py"),
    base_path.joinpath("markdown", "postprocessors.py"),
    base_path.joinpath("markdown", "serializers.py"),
    base_path.joinpath("markdown", "util.py"),
    base_path.joinpath("markdown", "htmlparser.py"),
    base_path.joinpath("markdown", "test_tools.py"),
    *sorted(base_path.joinpath("markdown", "extensions").rglob("*.py")),
]

for src_path in modules:
    path = src_path.relative_to(base_path)
    module_path = path.with_suffix("")
    doc_path = path.with_suffix(".md")
    full_doc_path = Path("reference", doc_path)

    parts = tuple(module_path.parts)

    if parts[-1] == "__init__":
        parts = parts[:-1]
        doc_path = doc_path.with_name("index.md")
        full_doc_path = full_doc_path.with_name("index.md")
    elif parts[-1].startswith("_"):
        continue

    nav_parts = [f"<code>{part}</code>" for part in parts]
    nav[nav_parts] = doc_path.as_posix()

    with mkdocs_gen_files.open(full_doc_path, "w") as fd:
        ident = ".".join(parts)
        fd.write(f"::: {ident}")
        if ident in per_module_options:
            yaml_options = yaml.dump({"options": per_module_options[ident]})
            fd.write(f"\n{textwrap.indent(yaml_options, prefix='    ')}")
        elif ident.startswith("markdown.extensions."):
            yaml_options = yaml.dump({"options": {"inherited_members": False}})
            fd.write(f"\n{textwrap.indent(yaml_options, prefix='    ')}")

    mkdocs_gen_files.set_edit_path(full_doc_path, ".." / path)

with mkdocs_gen_files.open("reference/SUMMARY.md", "w") as nav_file:
    nav_file.writelines(nav.build_literate_nav())
